(* ---------------------------------------------------------------------- *)
(* SLAM.m  v1.0 

SLAM is a Mathematica package that provides an interface between LesHouches  
spectrum generators (SoftSusy, SPheno, Suspect) and SQL data bases.

The interface allows the user to save calculated spectra in a data base and
recycle the existing spectra when already used input settings have been detected.



  Copyright (C) 2013 Peter Marquard & Nikolai Zerf

  This work is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported License. 
  To view a copy of this license, visit 
  http://creativecommons.org/licenses/by-sa/3.0/ 
  or send a letter to 
  Creative Commons, 
  444 Castro Street, Suite 900, 
  Mountain View, California, 94041, USA.

*)
(* ---------------------------------------------------------------------- *)

BeginPackage["SLAM`"];

(* ---------------------------------------------------------------------- *)
(* Usage Messages                                                         *)
ObtainLesHouchesSpectrum::usage="Generates a particle spectrum using a spectrum generator.
 Generated spectra can be saved to an SQL data base.
 ObtainLesHouchesSpectrum[] checks the SQL data base each time its called.
 When it finds matching input values in the SQL data base the stored spectrum is reused,
 instead of rerunning the corresponding spectrum generator.
 To see all possible Options use the command: #[[1]]&/@Options[ObtainLesHouchesSpectrum].";

ReadLesHouchesSpectrumFile::usage="Reads a file in LesHouches notation and returns the content in dependence of the Option InputRequest.
InputRequest->All: Returns the full file content in Mathematica notation.
                   All EVs are still contained as strings.
InputRequest->AllFormated: Same as above, but using the input of Option DataStructure to convert the EVs to proper formats like Real or Integer.
                           The Option DataStructure has to be of the format given in:
                                 DataStructure/.Options[ReadLesHouchesSpectrumFile]"

A0::usage="\!\(\*SubscriptBox[\(A\), \(0\)]\).Trilinear coupling soft susy breaking parameter at the GUT scale. Used for e.g. in the msugra scenario"; 
AdoptEntryFormats::usage="Can be True or False. 
When True the function FilterLesHouchesBlockToInputRequest[] converts EVs with the help of the 
 InputRequest list to their requested format (String,Integer,Real).
When False the EVs are kept as Strings.";
AllFormated::usage="Is a possible value for the Option InputRequest of the function ReadLesHouchesSpectrumFile[]. 
 If set the function will return the content of the LesHouches files in Mathematica notation using proper formats for EVs.";
AppendToFile::usage="Can be True or False.
When True WriteLesHouchesFile[] appends its output after the already existing content of the file.
When False WriteLesHouchesFile[] overwrites any already existing content in the file with its output.";
asmz::usage="\!\(\*SubscriptBox[\(\[Alpha]\), \(s\)]\) (\!\(\*SubscriptBox[\(M\),\(Z\)]\)). Input value for spectrum generators."; 
CheckList::usage="Is an Option of SelectSQLLesHouchesTable[]. It specifies the required values in the SQL data base, 
 which have to agree in order to find a matching data base entry."; 
ClearDataBase::usage="Can be False,True or Only.
 When False nothing happens.
 When True the relevant table is completly removed from the SQL data base and a new one is created.
 When Only the table gets only removed. No new table is created.";
DataStructure::usage="Is an Option of the function ReadLesHouchesSpectrumFile[].
 It can be:

 All
 AllFormated
 
 or list of nested rules like given by:

 DataStructure/.Options[ReadLesHouchesSpectrumFile]";
Direction::usage="Can be In or Out.
When reading LesHouches generator input files it should be In.
When reading LesHouches generator output files it should be Out.
Input and output files for generators can have the same names for blocks.
Due to the fact input AND output file is saved completly in the SQL data base, there is a need to differ between
the two sources. This is done with the help of an addition prefix:\"I_\" or \"O_\" in dependence of this Option."; 
ExtendDataBase::usage="Can be True or False.
When True new data is saved in the SQL data base.
When False no new data is save in the SQL data base."; 
GF::usage="\!\(\*SubscriptBox[\(G\), \(F\)]\). Fermi's coupling constant. Input value for spectrum generators."; 
InputFilePath::usage="Is a String which contains the full path to an LesHouches spectrum generator input file.";
InputList::usage="Is an Option of WriteLesHouchesFile and holds the definition of the input, 
which should be written to the file.
An example for such an input can be displayed with:

InputList /. Options[WriteLesHouchesFile]
";
InputRequest::usage="Allows the user to specify his data request in a defined Rule format.
The InputRequest tells ObtainLesHouchesSpectrum[], which LesHouches entries should be extracted from the spectrum generator output file.
In addition the users has to specify the format (String,Real,Integer) and Symbolic representation of the EV.
Use: InputRequest/.Options[ObtainLesHouchesSpectrum]
to get an example of the InputRequest format."; 
invalpha::usage="1/\[Alpha]. Input value for spectrum generators."; 
KeepSQLConnection::usage="Can be True or False.
When True the connection to the SQL data base is kept without any data transfer.
When False the connection to the SQL data is closed after each data transfer."; 
lambda::usage="\[CapitalLambda]. Breaking parameter in gauge mediated symmetry breaking scenario."; 
m0::usage="\!\(\*SubscriptBox[\(m\), \(0\)]\). Unified soft breaking parameters of scalars at the GUT scale.";
m12::usage="\!\(\*SubscriptBox[\(m\), \(1/2\)]\). Unified soft breaking parameters of gauginos at the GUT scale."; 
m32::usage="\!\(\*SubscriptBox[\(m\), \(3/2\)]\). Gravitino mass parameter. Used in the anomaly mediated symmetry breaking scenario."; 
M1::usage="\!\(\*SubscriptBox[\(M\), \(1\)]\). Soft mass breaking parameters of the U(1) gauginos.
Input value for the \"p19MSSM1\" and \"p19MSSM2\" low scale scenario.";
M3::usage="\!\(\*SubscriptBox[\(M\), \(3\)]\). Soft mass breaking parameters of the SU(3) gauginos.
Input value for the \"p19MSSM1\" low scale scenario.";
MA::usage="\!\(\*SubscriptBox[\(M\), \(A\)]\). Mass of the pseudoscalar Higgs-boson \!\(\*SuperscriptBox[\(A\), \(0\)]\). Parameter in the low scale scenarios."; 
MaxSQLConnectionAttempts::usage="Is of Integer type and by default 10.
When an SQL server is too busy to handle a connection attempt it might fail.
Thus it might be useful to try to connect later again. 
The maximal number of such connection attempts are defined with this parameter."; 
mb::usage="\!\(\*SubsuperscriptBox[\(m\), \(b\), OverscriptBox[\(MS\), \(_\)]]\) (\!\(\*SubscriptBox[\(m\), \(b\)]\)). \!\(\*OverscriptBox[\(MS\), \(_\)]\) bottom quark mass at the bottom quark mass. Input value for spectrum generators."; 
mmess::usage="Mass of the messenger particle in the gauge mediated symmetry breaking model."; 
Model::usage="Is of type String and defines the super symmetry breaking model.
Currently following model types are supported:

GUT scenarios:
\"msugra\" (minimal gravity mediated symmetry breaking).
\"mgmsb\" (gauge mediated symmetry breaking).
\"mamsb\" (anomaly mediated symmetry breaking).

Low scale scenarios:
\"gluophob\" (gluophobic higgs).
\"lightstau\" (light stau).
\"lightstaudeltatau\" (light stau including \!\(\*SubscriptBox[\(\[CapitalDelta]\), \(\[Tau]\)]\) corrections).
\"lightstop\" (light stop).
\"lowmh\" (low-\!\(\*SubscriptBox[\(M\), \(H\)]\)).
\"mhmax\" (mh max).
\"mhmaxup\" (updated mh max).
\"mhmod+\" (mh mod +).
\"mhmod-\" (mh mod -).
\"nomix\" (no mixing).
\"p19MSSM1\"(19 parameter MSSM plane I).
\"p19MSSM2\"(19 parameter MSSM plane II).
\"smallalpha\" (small effective alpha).
\"tauphobic\" (\[Tau]-phobic).

 pMSSM Benchmarks for Snowmass 2013:
\"401479\" (Bino-stop coannihilation).
\"1046838\" (Pure Higgsino).
\"2342344\" (Bino-squark coannihilation (Spoke 1 & 2)).
\"2387564\" (A funnel).
\"2750334\" (Well-tempered neutralino).
\"374345\" (Super-WIMP)."; 
MtOS::usage="\!\(\*SubscriptBox[\(M\), \(t\)]\). On-Shell top quark mass. Input value for spectrum generators."; 
Mtau::usage="\!\(\*SubscriptBox[\(M\), \(\[Tau]\)]\).On-Shell tau mass. Input value for spectrum generators."; 
mu::usage="\[Mu]. Bilinear coupling constant of \!\(\*SubscriptBox[\(H\), \(u\)]\) and \!\(\*SubscriptBox[\(H\), \(d\)]\). Parameter in the low scale scenario \"lowmh\".";
muRen::usage="\!\(\*SubscriptBox[\(\[Mu]\), \(R\)]\). Renormalization scale. Scale where \!\(\*OverscriptBox[\(DR\), \(_\)]\) spectrum parameters are evaluated.";
Mw::usage="\!\(\*SubscriptBox[\(M\), \(W\)]\). Mass of the W-Boson."; 
Mz::usage="\!\(\*SubscriptBox[\(M\), \(Z\)]\). Mass of the Z-Boson.";
mslepton::usage="\!\(\*SubscriptBox[\(m\), OverscriptBox[\(l\), \(~\)]]\) = \!\(\*SubscriptBox[\(m\), SubscriptBox[OverscriptBox[\(e\), \(~\)],\(R\)]]\) = \!\(\*SubscriptBox[\(m\), SubscriptBox[OverscriptBox[\(e\), \(~\)], \(L\)]]\).
Common left-/right-selectron soft SUSY breaking mass term.
Input value for the \"p19MSSM2\" low scale scenario.";
nmess::usage="Number of minimal copies of the messenger sector in the gauge mediated symmetry breaking model.";
OutputFilePath::usage="Is a String which contains the full path to an LesHouches spectrum generator output file."; 
OutputFormat::usage="Defines the format of the output. Possible Options are:
 LesHouchesBlockRule
   Results are represented in the structure of the SUSY LesHouches accord using Mathematica synthax. 
 TableDefinitionList
   Returns the Result in 2 column table.
   The first column contains the name of the corresponding SQL column as string.
   The second column contains the value entry stored in the corresponding SQL column.
   Example:
            {{\"MASS_1000001\",323.0},{\"MASS_1000001_COMMENT\",\" ~d_L\"}}
 PureValueList
   Returns the pure results in list format.
   Example:
            {323.0,\" ~d_L\"}
   ";
SLAMPrint::usage="SLAMPrint is given by the Print function, except $SLAMSendPrintToFile=True.
Then all Print statement are sent to an output file.
This should be used for debugging, only!";
$SLAMSendPrintToFile::usage="Can be True or False. 
When True all SLAMPrint statements are sent to an output file.
When False all SLAMPrint statements are displayed using the Print function.";
LesHouchesBlockRule::usage="Is a possible Option Value of OutputFormat. See ?OutputFormat for more information.";
$LesHouchesSQLConnectionList::usage="Contains a list of all opened SQL connections.";
LoadCompleteSpectra::usage="Can be True or False.
When True, the full spectrum data is loaded from the data base.
Even, when the user requested only a part of the data.
When False, only the data requested from the user gets loaded from the data base.
This can reduce the loading times, if only a few data values are needed from the spectra.";
TableDefinitionList::usage="Is a possible Option Value of OutputFormat. See ?OutputFormat for more information.";
PureValueList::usage="Is a possible Option Value of OutputFormat. See ?OutputFormat for more information.";
OutputRequestList::usage="Specifies the output to be read from the SQL data base in a list.
It has to be a list of existing SQL data base column names.
Example:
{\"O_MINPAR_4\", \"O_MINPAR_4_COMMENT\", \"O_MINPAR_5\", 
\"O_MINPAR_5_COMMENT\", \"O_MASS_24\", \"O_MASS_24_COMMENT\", \"O_MASS_25\", 
\"O_MASS_25_COMMENT\", \"O_MASS_35\", \"O_MASS_35_COMMENT\", \"O_MASS_36\", 
\"O_MASS_36_COMMENT\", \"O_MASS_37\", \"O_MASS_37_COMMENT\", 
\"O_MASS_1000001\", \"O_MASS_1000001_COMMENT\"}"; 
RefreshDataBase::usage="Can be True or False.
When True old existing results in SQL data base are overwritten by new ones.
That means the calculation is performed although there is a matching result in data base.
When False stored results in the SQL data are used and kept.
No new calculation is performed when a matching result exists in data base.";
RemoveTemporaryFiles::usage="Can be True or False.
When True all temporary created files are deleted after use.
When False all temporary created files are kept on disk.";
SecureEntryFormats::Usage="Can be True or False. 
When True ensures that no rounding errors accure when using direct model input.
When False the double conversion behind is not used.
It is highly recommendet to leave this Option untouched!";
signmu::usage="Sign[\!\(\*SubscriptBox[\(\[Mu]\), \(SUSY\)]\)]. Sign of the soft breaking term ~\!\(\*SubscriptBox[\(H\), \(u\)]\)\!\(\*SubscriptBox[\(H\), \(d\)]\). 
Input parameter in SUSY breaking scenarios.";
Silent::usage="Can be True or False.
If True ObtainLesHouchesSpectrum will not print processing messages."; 
SpectrumGenerator::usage="Defines the the spectrum generator to be used for the production of particle spectra.
Results of different spectrum generators are stored independently in the SQL data base.
Currently supported spectrum generators/Options are:
\"softsusy\"
\"spheno\"
\"suspect\"."; 
SQLDataBase::usage="Is of String type and contains the address to the SQL data base.
When you are using  SQLDataBaseType->\"hsqldb\" the address is just a path where the data base files get stored by Mathematica.
 Example: 
 \"/x1/Project_ggh/common/numerics/spectra/sql/database\". 
When you are using a MySQL server and for example setting SQLDataBaseType->\"MySQL(Connector/J)\" this Option holds the address
 including the port for the MySQL server.
 Example:
 \"ttpearth.particle.uni-karlsruhe.de:30405\""; 
SQLDataBaseOptions::usage="Defines the Options of the OpenSQLConnection[] function (see Mathematica help for more information).
Example:
{\"Name\" -> \"Spectrum\", \"Username\" -> \"zerf\"}."; 
SQLDataBaseType::usage="Is of String type and defines the type of data base to be used. 
It is needed in the OpenSQLConnection[] function (see Mathematica help for more information).
By default you are using:
\"hsqldb\".
  This means no real data base server is running, but Mathematica is hosting a server for you on your current computer.
In case you have a MySQL server hosting your data base, this Option might look like:
\"MySQL(Connector/J)\"."; 
SQLVarCharLength::usage="Is of Integer type and defines the maximum length of all VARCHAR declared in the data base."; 
TableDeclarationList::usage="Is a list which is used during the declaration of SQL data base table.
This list is only need when no matching table was found in data base.
It needs to be declared only once before data can be written into it.
Example of the required format:
{\"I_SMINPUTS_INFO\", \"VARCHAR\"}, {\"I_MINPAR_COMMENT\", \"VARCHAR\"}, {\"I_MINPAR_INFO\", \"VARCHAR\"}, 
{\"I_EXTPAR_COMMENT\", \"VARCHAR\"}, {\"I_EXTPAR_INFO\", \"VARCHAR\"}, {\"O_SPHENOINFO_1\", \"INTEGER\"}, 
{\"O_SPHENOINFO_1_COMMENT\", \"VARCHAR\"},{\"O_SPHENOINFO_2\", \"INTEGER\"}, {\"O_SPHENOINFO_2_COMMENT\", \"VARCHAR\"}.
Currently used data formats are:
\"VARCHAR\"
\"INTEGER\"
\"DOUBLE\"."; 
TableName::usage="Is of (upper case) String type and defines the name of the Table used in the SQL data base."; 
tanbeta::usage="tan(\[Beta]). Ratio of the two Higgs vacuum expectation values. 
tan(\[Beta])=\!\(\*SubscriptBox[\(v\), \(u\)]\)/\!\(\*SubscriptBox[\(v\), \(d\)]\).";
UseDataBase::usage="Can be True or False.
When True the SQL data base is checked for matching result before calculating it.
When False the SQL data base is not checked or extended at all.
 Everything is calculated from \"scratch\". The data base is not used."; 
WorkingFolder::usage="Can be Automatic or a String containing a path where temporary files can be treated.
It is recommended to keep the default Automatic setting.";
WriteLesHouchesFile::usage="Writes a SLHA file at location of option InputFilePath with the content provided in the option InputList.
An example of the format required for the value of InputList can be printed with the command:

InputList /. Options[WriteLesHouchesFile]";
xBFS2013::usage="Free parameter of Spoke 1 in Bino-squark coannihilation scenario \"2342344\".
xBFS2013=yBFS2013=0 corresponds to the point \"2342344\".";
yBFS2013::usage="Free parameter of Spoke 2 in Bino-squark coannihilation scenario \"2342344\".
yBFS2013=yBFS2013=0 corresponds to the point \"2342344\".";

OSQLR::usage="Basic idea:
  OSQLR[f][x] checks the SQL data base for matching numeric values of x.
  If the value of x is already contained, it loads the result of f[x] directly from the data base.
  Only if the value of x is not contained in the data base, the (possibly involved) calculation of f[x] is performed.
  OSQLR can be used to avoid time consuming calculations, which appear more than a single time, especially beyond the runtime of a single Mathematica kernel.
  That means results are kept, even if one quits the Mathematica kernel.

Usage requirements:
  OSQLR can be used for any function f which fulfills the following conditions:
  -The arguments or OptionValues of the function must be of the following type:
   String
   Real
   Integer.
  -The result of the function f can be composed from following basic types:
   String
   Real
   Integer.

Usage:
  In case where the function f does only have optional arguments that means it's definition looks like:

       Options[f]={x->1};
       f[opt:OptionsPattern[]]:= OptionValue[x];
  
  OSQLR will outomatically detect that the parameter x is of Integer type.
  That means if you run OSQLR on the function f for the first time you only need to define the format of the expected result as follows:

       OSQLR[f,
             ResultStructure->1,
             ResultTypes->{1 -> SQLDataTypeName[Integer]}
            ][x->2]

  ResultStructure->1 tells that the result of f[x-> 2] is just one Value,
  where
  ResultTypes->{1 -> SQLDataTypeName[Integer]}
  tells OSQLR that this Value is of Type Integer.
  If you have run this command once, OSQLR will remember the properties of function f for the rest of the session.
  It automatically will create the SQL data base table and manage the data transfer.

  In case the function f does in addition depend on a regular argument for e.g.:

       Options[f]={x->1};
       f[a_,opt:OptionsPattern[]]:= {OptionValue[x],a};
 
  You need to specify the type of the argument a as follows:

       OSQLR[f,
               ArgumentTypes -> {1 -> SQLDataTypeName[Real]},
               ResultStructure->{1,2},
               ResultTypes->{1-> SQLDataTypeName[Integer],2-> SQLDataTypeName[Real]}
             ][1.0,x->2]  

  In addition the structure of the result ({OptionValue[x],a}) has changed and has to be adjusted in the ResultStructure Option.
  Once you have run this command OSQLR will modify its own Option for FunctionSettings.
  It will append all information needed for the interface to the SQL data base.
  Use 
        \"f\"/.(FunctionSettings/.Options[OSQLR])

  to see all relevant properties.

Note:
  When the definitions of the function f changes during a Mathematica session, 
  you have to remove the existing data base table and the corresponding FunctionSetting before using OSQLR again on this function.
  This can be done by calling OSQLR with the Option: 
  ClearFunctionStorage -> Only."



FunctionSettings::usage="Option of OSQLR where the interface settings of all analyzed functions are saved.";
ClearFunctionStorage::usage="Can be True or False.
When True all saved settings of a function in the OSQLR Option FunctionSettings are removed and the SQL data base table is cleared.
When False nothing happens.";
ArgumentTypes::usage="Is a List which defines the types of non optional arguments.
Example:
{1 -> SQLDataTypeName[Real],2 -> SQLDataTypeName[Integer],3->SQLDataTypeName[Double]}.
It Means:
The first non optional argument is of type Real,
the second /     /     argument is of type Integer,
the third  /     /     argument is of type Double.";
ResultTypes::usage="Is a List which defines the types of result values.
Example:
{1 -> SQLDataTypeName[Real],2 -> SQLDataTypeName[Integer],3->SQLDataTypeName[Double]}.
It Means:
The first  result value is of type Real,
the second result value is of type Integer,
the third  result value is of type Double."; 
ResultStructure::usage="Defines the structure of the result composed from result values.
Example 1:
ResultStructure->1
It Means:
The corresponding function returns a single result value.

Example 2:
ResultStructure->{1,2,3}
It Means:
The corresponding function returns a list with 3 result values.";
SQLDataTypeName::usage="Function which transforms the Mathematica types:
String,
Integer,
Real 
in its corresponding SQL data types:
\"VARCHAR\",
\"INTEGER\",
\"DOUBLE\".";


EntryKey::usage="LesHouches key entry in a LesHouches-Convention file.";

EntryValue::usage="LesHouches value entry in a LesHouches-Convention file.";
EntryComment::usage="LesHouches comment entry in a LesHouches-Convention file.";
BlockComment::usage="LesHouches block comment in a LesHouches-Convention file.";
BlockInfo::usage="LesHouches block info in a LesHouches-Convention file.";


EV::usage="LesHouches value entry in a LesHouches-Convention file.";
EC::usage="LesHouches comment entry in a LesHouches-Convention file.";
BC::usage="LesHouches block comment in a LesHouches-Convention file.";
BI::usage="LesHouches block info in a LesHouches-Convention file.";



$MissingData::usage="Is used as place holder when ObtainLesHouchesSpectrum[] could not find a machting value in the obtained output."
(* ---------------------------------------------------------------------- *)
(* Error Messages                                                         *)


SLAM::UnknownSpecGen="Spectrum generator is not ready or unknown!";
SLAM::NoSpecGenReady="No spectrum generator is currently ready to use!\nMake sure that you have at least one spectrum generators properly installed!";
SLAM::UnknownModel="Requested model `1` is unknown!";
SLAM::NoFileMove="Could not move temporary files to working directory!";
SLAM::SpecGenNonZeroExit="Spectrum generator exited with non-zero exit status `1`.";
SLAM::SpecGenNonZeroExitPlusMessage="Spectrum generator exited with non-zero exit status `1`\nincluding message:\n`2`.";
SLAM::SpecGenMessage="Spectrum generator `1` sent message:\n `2`";
SLAM::SpecGenError="Spectrum generator wrote error entry:\n `1`";
SLAM::SpecGenWarning="Spectrum generator wrote warning entry:\n `1`";
SLAM::SedError="Sed failed to remove pure comment lines!";
SLAM::NoDataBaseDirectory="Cannot access directory for Mathematica internal data base.\nTrying to create temporary directory.\nIf this fails make sure path given in\n SQLDataBaseValue (`1`)\n exists and is accessible!";
SLAM::NoTmpInputDirectory="Cannot access directory for temporary input files.\nTrying to create temporary directory.\nIf this fails make sure path given in\n InputFilePathValue (`1`)\n exists and is accessible!";
SLAM::NoTmpOutputDirectory="Cannot access directory for temporary output files.\nTrying to create temporary directory.\nIf this fails make sure path given in\n OutputFilePathValue (`1`)\n exists and is accessible!";
SLAM::NoFileRemove="Could not remove temporary files!";
SLAM::NoLHAFileRead="Could not read LesHouches file!";
SLAM::NoDBConnect="Could not connect to data base!";
SLAM::IllegalDataRequest="You ordered data not present in data base. Check your InputRequest!";
SLAM::NoModselInModel="Could not find a \"MODSEL\" block in Your model. Check if the Model definition is correct!";
SLAM::MultipleDBEntry="Found more than one entry in data base matching all input definitions!\nTaking first entry...";
SLAM::NoResults="Could not obtain proper result with: `1`";
SLAM::WrongResultFormat="Output result values of `1` do not match given data type!";


(* ---------------------------------------------------------------------- *)
Begin["`Private`"];

(* Older versions used long variable name, which is not that comfortable when writing input...*)
EntryValue=EV;
EntryComment=EC;
BlockComment=BC;
BlockInfo=BI;


(* Loading config file *)
Get[StringReplace[$Input, {".m" -> ".config.m"}]];



(* Sends all Print statements to the file below if turned  on. Used for debugging only!*)
$SLAMSendPrintToFile=False;

If[!$SLAMSendPrintToFile,
  SLAMPrint= Print;,

  PrintFilePathName = DirectoryName[Global`InputFilePathValue]<>"SLAM.print.out";
  WriteToPrintFile := (PrintFile = OpenAppend[PrintFilePathName]; WriteString[PrintFile,(StringJoin @@ (ToString /@ {##})) <> "\n"]; Close[PrintFile];) &;
  SLAMPrint = WriteToPrintFile;

  SendMessageToPrintFile := If[Length[Cases[{##}, $Off[], {0, Infinity}]] === 0, 
                           (PrintFile = OpenAppend[PrintFilePathName]; 
                           WriteString[PrintFile, "WARNING: " <> ToString[{##}] <> "\n"];
                           Close[PrintFile];)] &;
  Unprotect[Message];
  Message = SendMessageToPrintFile;

  Print["Writing all Print statements and Messages to: ",PrintFilePathName];
];


(* Checking if spectrum generators are properly installed *)
If[(FileType[SLAM`Private`ProgramPath[#]] === File) ,
   InstalledQ[#] = True;SLAMPrint[#<>" is ready."];,
   InstalledQ[#] = False;SLAMPrint[#<>" is not ready."];] & /@ {"suspect", "spheno", "softsusy"};

If[ !(Or @@(InstalledQ[#]&/@ {"suspect", "spheno", "softsusy"})),Message[SLAM::NoSpecGenReady];];

(* Checking if the temporary directories are reachable *)
If[FileType[DirectoryName[Global`InputFilePathValue]]=!=Directory,
   Message[SLAM::NoTmpInputDirectory,Global`InputFilePathValue];
   CreateDirectory[DirectoryName[Global`InputFilePathValue]];
];

If[FileType[DirectoryName[Global`OutputFilePathValue]]=!=Directory,
   Message[SLAM::NoTmpOutputDirectory,Global`OutputFilePathValue];
   CreateDirectory[DirectoryName[Global`OutputFilePathValue]];
];

(* Checking if install default Mathematica data base directory is reachable -> NOT NEEDED IS DONE AUTOMATICALLY *)
(*If[Global`SQLDataBaseValue==="~/.SLAM/database",
   If[FileType[DirectoryName[Global`SQLDataBaseValue]]=!=Directory,
      Message[SLAM::NoDataBaseDirectory,Global`SQLDataBaseValue];
      CreateDirectory[DirectoryName[Global`SQLDataBaseValue]];
     ];
];*)


(* 
 Function which hands down options from master to subfunctions.
 If no Options are specified in call of master function the default options of the master function is traded to the subfunction.
*)
(*
Old and buggy version!
TradeOptions[MainFunction_, SubFunction_, opts___] := ((*SLAMPrint["Trading Rule: "<>ToString[MainFunction]<>"->"<>ToString[SubFunction]];*)Sequence @@ (Table[Options[SubFunction][[i, 1]] -> (Options[SubFunction][[i, 1]] /. FilterRules[{opts}, Options[SubFunction]] /. FilterRules[Options[MainFunction], Options[SubFunction]]), {i, 1, Length[Options[SubFunction]]}]/. Rule -> Rul /. (Rul[a_, a_] -> Sequence[]) /. Rul -> Rule))
*)

TradeOptions[MainFunction_, SubFunction_, opts___] := Sequence @@ (Table[Options[SubFunction][[i, 1]] -> (If[# === Options[SubFunction][[i, 1]],
                                                                                                             # /. FilterRules[Options[MainFunction], Options[SubFunction]],
                                                                                                             #] &@(Options[SubFunction][[i, 1]] /. FilterRules[{opts}, Options[SubFunction]])), {i, 1, Length[Options[SubFunction]]}] /. Rule -> Rul /.Rul[a_, a_] -> Sequence[] /. Rul -> Rule)




(* Mathematica 6.0 and below 
do not support Gather[] function 
  -> Implementing MyGather[] function
do not support FileNameTake[] function
  -> Implementing MyFileNameTake[] function
*)
If[$VersionNumber<7.0,
   MyGather[expr_, comp_] := Module[{GroupeList, CurrentElement},
   
         GroupeList = {};
   
         Return[(Reap[((If[Length[GroupeList] === 0,
              AppendTo[GroupeList, #]; Sow[#, #];,
              CurrentElement = #;
              
              If[! (Or @@ (comp[CurrentElement, #] & /@ GroupeList)),
               AppendTo[GroupeList, CurrentElement];
               Sow[CurrentElement, CurrentElement];,
               If[(comp[CurrentElement, #]),
                   Sow[CurrentElement, #];
                   ] & /@ GroupeList;
               ];
              ];) & /@ expr);, _][[2]])]
          ];

   MyFileNameTake[X_String] := StringReplace[X, RegularExpression[".*[/]([^/]*$)"] :> "$1"];

   Gather=MyGather;
   FileNameTake=MyFileNameTake;
]


FORTRANInputForm[x_] := StringReplace[ToString[ScientificForm[NumberForm[N[x], {9, 8}, NumberMultiplier -> " "], 9, NumberMultiplier -> " "]], 
{RegularExpression["\\s*[-](\\d\\d)\n([-]*\\d+[.]\\d+)\\s+\\d+"] :> "$2E-$1", 
   RegularExpression["\\s*[-](\\d)\n([-]*\\d+[.]\\d+)\\s+\\d+"] :> "$2E-0$1", 
   RegularExpression["\\s*(\\d\\d)\n([-]*\\d+[.]\\d+)\\s+\\d+"] :> "$2E+$1", 
   RegularExpression["\\s*(\\d)\n([-]*\\d+[.]\\d+)\\s+\\d+"] :> "$2E+0$1", 
   RegularExpression["^([+]*[-]*\\d[.]\\d+)$"] :> "$1E+00", __ -> "This number format is not supported!"}]

MATHEMATICAInputForm[x_String] := ToExpression[StringReplace[x, RegularExpression["([+-]*\\d+)E([+-]*\\d+)"] :> "$1*10^($2)"]]

Options[WriteSpectrumGeneratorInput] = {
   InputFilePath -> Global`InputFilePathValue,
   SpectrumGenerator -> "spheno",(* "softsusy" *)
   Model -> "msugra",(*"mgmsb","mamsb","mhmax","nomix","gluophob","smallalpha"*)
   (* Renormalisation scale for the DRbar parameters. *)         
   muRen -> Global`muRenValue,
   (* SMINPUTS.                                   *)
   invalpha -> Global`invalphaValue,(* 1/alpha(mu=Mz) MSbar*)
   GF -> Global`GFValue,
   asmz -> Global`asmzValue,(* alphas(Mz) MSbar*)
   Mz -> Global`MzValue,
   Mw -> Global`MwValue,
   mb -> Global`MbMSbarValue,(* mb(mb) MSbar)*)
   MtOS -> Global`MtOSValue,(* Mt Pole*)
   Mtau -> Global`MtauValue,(* mtau Pole*)
   (* msugra                                      *)   
   m0 -> Global`m0Value,
   m12 -> Global`m12Value,
   tanbeta -> Global`tanbetaValue,
   signmu -> Global`signmuValue,
   A0 -> Global`A0Value,
   (*  mgmsb                                      *)         
   lambda -> Global`lambdaValue,
   mmess -> Global`mmessValue,
   nmess -> Global`nmessValue,
   (*  mamsb                                      *)
   m32 -> Global`m32Value,
   (* for benchmark points of hep-ph/0202167      *)
   MA -> Global`MAValue,
   M1 -> Global`M1Value,
   M3 -> Global`M3Value,
   mslepton-> Global`msleptonValue,
   (* for benchmark scenario lowmh defined in hep-ph/1302.7033   *)
   mu-> Global`muValue,
   (* for pMSSM Benchmarks for Snowmass 2013 (http://www.slac.stanford.edu/~aismail/snowmass/index.html) *)
   xBFS2013->Global`xBFS2013Value,
   yBFS2013->Global`yBFS2013Value
};


WriteSpectrumGeneratorInput[opt : OptionsPattern[]] := Module[{InputFileString,WriteLineFunction},
    

    If[OptionValue[InputFilePath]===String,
       InputFileString="";
       WriteLineFunction:=(InputFileString=StringJoin[##])&;,
       InputFileString = OpenWrite[OptionValue[InputFilePath]];
       WriteLineFunction=WriteString;
    ];


    
    (*Block Modsel*)
    
    WriteLineFunction[InputFileString, 
     "# SLHA input file generated by SLAM.m\n", 
     "Block MODSEL\n"];
    

    
    (*Only the following generators are supported *)
    Which[
     OptionValue[SpectrumGenerator]==="softsusy" && InstalledQ["softsusy"],
     Hold;,
     OptionValue[SpectrumGenerator]==="spheno" && InstalledQ["spheno"],
     Hold;,
     OptionValue[SpectrumGenerator]==="suspect" && InstalledQ["suspect"],
     Hold;,
     True,
     Message[SLAM::UnknownSpecGen];
     Abort[];];
    
    
    If[NumberQ[OptionValue[muRen]] && (OptionValue[SpectrumGenerator]=!="suspect"), (* suspect cannot deal with renormalisation scale entry*)
       WriteLineFunction[InputFileString, 
                   " 11   1   # One value of the renormalisation scale\n", 
                   " 12   ", FORTRANInputForm[OptionValue[muRen]],   "  # renormalisation scale\n"]];

   (*Specification of model *)
    Switch[OptionValue[Model], 
           "msugra", 
           WriteLineFunction[InputFileString, " 1   1    # msugra\n"];, 
           "mgmsb", 
           WriteLineFunction[InputFileString, " 1   2    # mgmsb\n"];,
           "mamsb", 
           WriteLineFunction[InputFileString, " 1   3    # mamsb\n"];, 
           "mhmax", 
           WriteLineFunction[InputFileString, " 1   0    # general (mhmax)\n"];,
           "mhmaxup", 
           WriteLineFunction[InputFileString, " 1   0    # general (mhmax updated)\n"];,
           "mhmod+", 
           WriteLineFunction[InputFileString, " 1   0    # general (mhmod+)\n"];,
           "mhmod-", 
           WriteLineFunction[InputFileString, " 1   0    # general (mhmod-)\n"];,
           "lightstop", 
           WriteLineFunction[InputFileString, " 1   0    # general (lightstop)\n"];,
           "lightstau", 
           WriteLineFunction[InputFileString, " 1   0    # general (lightstau)\n"];,
           "lightstaudeltatau", 
           WriteLineFunction[InputFileString, " 1   0    # general (lightstau including deltatau corr.)\n"];,
           "tauphobic", 
           WriteLineFunction[InputFileString, " 1   0    # general (tauphobic)\n"];,
           "lowmh", 
           WriteLineFunction[InputFileString, " 1   0    # general (lowmh)\n"];, 
           "nomix", 
           WriteLineFunction[InputFileString, " 1   0    # general (no-mixing)\n"];, 
           "gluophob",
           WriteLineFunction[InputFileString, " 1   0    # general (gluophobic higgs)\n"];, 
           "smallalpha", 
           WriteLineFunction[InputFileString, " 1   0    # general (small effective mixing angle)\n"];, 
           "p19MSSM1", 
           WriteLineFunction[InputFileString, " 1   0    # general (19 parameter MSSM plane 1)\n"];, 
           "p19MSSM2", 
           WriteLineFunction[InputFileString, " 1   0    # general (19 parameter MSSM plane 2)\n"];,
           "401479", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 401479 (Bino-stop coannihilation)\n"];,
           "1046838", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 1046838 (Pure Higgsino))\n"];,
           "2342344", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 2342344 (Bino-squark coannihilation))\n"];,
           "2387564", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 2387564 (A funnel))\n"];,
           "2750334", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 2750334 (Well-tempered neutralino))\n"];,
           "374345", 
           WriteLineFunction[InputFileString, " 1   0    # general (pMSSM BFS2013 374345 (Super-WIMP))\n"];,
           _,
           Message[SLAM::UnknownModel,OptionValue[Model]];Abort[];];

     (*Sepctrum generator dependend blocks*)
     Switch[OptionValue[SpectrumGenerator],
            "softsusy", 
            WriteLineFunction[InputFileString, 
            "Block SOFTSUSY\n", 
            " 5   1   # two-loop RGEs for scalars?\n"];,
            "spheno", 
            WriteLineFunction[InputFileString, 
            "Block SPhenoInput\n", 
            " 11   0   # don't calculate branching ratios\n", 
            " 21   0   # don't calculate cross-sections\n"];,
            "suspect",
            WriteLineFunction[InputFileString, 
            "Block SU_ALGO\n", 
            " 2   21   # 2-loop RGE (defaut, 1-loop RGE is: 11 instead) \n"];];

            WriteLineFunction[InputFileString, 
            "Block SMINPUTS\n", 
            " 1   ", FORTRANInputForm[OptionValue[invalpha]], "  # alpha^(-1) SM MSbar(Mz)\n", 
            " 2   ", FORTRANInputForm[OptionValue[GF]], "   # G_Fermi\n", 
            " 3   ", FORTRANInputForm[OptionValue[asmz]], "   # alpha_s(Mz) SM MSbar\n", 
            " 4   ", FORTRANInputForm[OptionValue[Mz]], "   # Mz(pole)\n", 
            " 5   ", FORTRANInputForm[OptionValue[mb]], "   # mb(mb) SM MSbar\n", 
            " 6   ", FORTRANInputForm[OptionValue[MtOS]], "   # Mt SM pole\n", 
            " 7   ", FORTRANInputForm[OptionValue[Mtau]], "   # mtau pole\n"];

    (* Model dependend inputs *)
    Switch[OptionValue[Model], 
    "msugra", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 1   ", FORTRANInputForm[OptionValue[m0]], "   # m_0\n", 
       " 2   ", FORTRANInputForm[OptionValue[m12]], "   # m_1/2\n", 
       " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "  # tan(beta)\n", 
       " 4   ", FORTRANInputForm[OptionValue[signmu]], "   #sign(mu)\n", 
       " 5   ", FORTRANInputForm[OptionValue[A0]], "   # A0\n"];
       If[OptionValue[SpectrumGenerator]==="suspect", 
          WriteLineFunction[InputFileString,"Block EXTPAR\n"];];,(* suspect needs this block line to work properly, although no information is given in that block...*) 
    "mgmsb", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 1   ", FORTRANInputForm[OptionValue[lambda]], "   # lambda\n", 
       " 2   ", FORTRANInputForm[OptionValue[mmess]], "   # M_mess\n", 
       " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       " 4   ", FORTRANInputForm[OptionValue[signmu]], "   # sign(mu)\n", 
       " 5   ", FORTRANInputForm[OptionValue[nmess]], "   # N_mess\n"];
       If[OptionValue[SpectrumGenerator]==="suspect", 
          WriteLineFunction[InputFileString,"Block EXTPAR\n"];];,(* suspect needs this block line to work properly, although no information is given in that block...*) 
    "mamsb", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 1   ", FORTRANInputForm[OptionValue[m0]], "   # m_0\n", 
       " 2   ", FORTRANInputForm[OptionValue[m32]], "   # m_3/2\n", 
       " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       " 4   ", FORTRANInputForm[OptionValue[signmu]], "   # sign(mu)\n"];
       If[OptionValue[SpectrumGenerator]==="suspect", 
          WriteLineFunction[InputFileString,"Block EXTPAR\n"];];,(* suspect needs this block line to work properly, although no information is given in that block...*) 
    "mhmax", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.  # M2\n", 
       " 3   ", FORTRANInputForm[0.8*1000.], "   # M3\n", 
       " 11   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   200.   # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   1000.   # Msusy\n", 
       " 32   1000.   # Msusy\n", 
       " 33   1000.   # Msusy\n", 
       " 34   1000.   # Msusy\n", 
       " 35   1000.   # Msusy\n", 
       " 36   1000.   # Msusy\n", 
       " 41   1000.   # Msusy\n", 
       " 42   1000.   # Msusy\n", 
       " 43   1000.   # Msusy\n", 
       " 44   1000.   # Msusy\n", 
       " 45   1000.   # Msusy\n", 
       " 46   1000.   # Msusy\n", 
       " 47   1000.   # Msusy\n",
       " 48   1000.   # Msusy\n", 
       " 49   1000.   # Msusy\n"],
    "mhmaxup", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.  # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[Sqrt[6]*1000. + 200./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   200.   # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   1000.   # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   1000.   # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1000.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1000.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1000.   # msbR\n"], 
    "mhmod+", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.  # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[1.6*1000. + 200./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[1.6*1000. + 200./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[1.6*1000. + 200./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   200.   # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   1000.   # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   1000.   # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1000.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1000.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1000.   # msbR\n"], 
    "mhmod-", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.  # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[-2.2*1000. + 200./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[-2.2*1000. + 200./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[-2.2*1000. + 200./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   200.    # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   1000.   # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   1000.   # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1000.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1000.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1000.   # msbR\n"], 
    "lightstop", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", 
       " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n",(* Note: Msusy=500. but this is too low for input scale (->tachyonic particles ) thats why we put 1000. In agreement with OS masses for st1 when comparing softsusy and feynhiggs  *)
       " 1   ",  FORTRANInputForm[350.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   350.    # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[2.2*500. + 350./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[2.2*500. + 350./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[2.2*500. + 350./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   350.    # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   1000.   # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   1000.   # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   500.    # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   500.    # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   500.    # msbR\n"],
    "lightstau", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.    # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[1.7*1000. + 500./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[1.7*1000. + 500./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   0.      # Atau\n", 
       " 23   500.    # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   245.    # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   245.    # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1000.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1000.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1000.   # msbR\n"],
    "lightstaudeltatau", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1000.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[400.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2    400.    # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[1.7*1000. + 450./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[1.7*1000. + 450./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   0.      # Atau\n", 
       " 23   450.    # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   250.    # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   250.    # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1000.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1000.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1000.   # msbR\n"],
    "tauphobic", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1500.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2    200.    # M2\n", 
       " 3   ", FORTRANInputForm[1500.], "   # M3\n", 
       " 11   ", FORTRANInputForm[2.9*1500. + 2000./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[2.9*1500. + 2000./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[2.9*1500. + 2000./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   2000.    # mu\n", 
       " 26 ", FORTRANInputForm[OptionValue[MA]], "  # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   500.    # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   500.    # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1500.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1500.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1500.   # msbR\n"],
    "lowmh", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ",  FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", " 0 ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+1500.^2]], "   # input scale\n", 
       " 1   ",  FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2    200.    # M2\n", 
       " 3    1500.    # M3\n", 
       " 11   ", FORTRANInputForm[2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta]], "   # Atau\n", 
       " 23   ", FORTRANInputForm[1.0*OptionValue[mu]], "    # mu\n", 
       " 26   110.     # MA\n", 
       " 31   500.    # mseL\n", 
       " 32   500.    # msmuL\n", 
       " 33   1000.   # mstauL\n", 
       " 34   500.    # mseR\n", 
       " 35   500.    # msmuR\n", 
       " 36   1000.   # mstauR\n", 
       " 41   1500.   # msqLg1\n", 
       " 42   1500.   # msqLg2\n", 
       " 43   1500.   # msqLg3\n", 
       " 44   1500.   # msuR\n", 
       " 45   1500.   # mscR\n", 
       " 46   1500.   # mstR\n", 
       " 47   1500.   # msdR\n",
       " 48   1500.   # mssR\n", 
       " 49   1500.   # msbR\n"], 
    "nomix", 
       WriteLineFunction[InputFileString, 
       "Block MINPAR\n", 
       " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
       "Block EXTPAR\n", 
       " 0   ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+2000.^2]], "   # input scale\n", 
       " 1   ", FORTRANInputForm[200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
       " 2   200.   # M2\n", 
       " 3   ", FORTRANInputForm[0.8*2000.], "   # M3\n", 
       " 11   ", FORTRANInputForm[200./OptionValue[tanbeta]], "   # At\n", 
       " 12   ", FORTRANInputForm[200./OptionValue[tanbeta]], "   # Ab\n", 
       " 13   ", FORTRANInputForm[200./OptionValue[tanbeta]], "   # Atau\n", 
       " 23   200.   # mu\n", 
       " 26   ", FORTRANInputForm[OptionValue[MA]], "   # MA\n", 
       " 31   2000.   # Msusy\n", 
       " 32   2000.   # Msusy\n", 
       " 33   2000.   # Msusy\n", 
       " 34   2000.   # Msusy\n", 
       " 35   2000.   # Msusy\n", 
       " 36   2000.   # Msusy\n", 
       " 41   2000.   # Msusy\n", 
       " 42   2000.   # Msusy\n", 
       " 43   2000.   # Msusy\n", 
       " 44   2000.   # Msusy\n", 
       " 45   2000.   # Msusy\n", 
       " 46   2000.   # Msusy\n", 
       " 47   2000.   # Msusy\n", 
       " 48   2000.   # Msusy\n", 
       " 49   2000.   # Msusy\n"], 
    "gluophob", 
      WriteLineFunction[InputFileString, "Block MINPAR\n", 
      " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
      "Block EXTPAR\n", 
      " 0   ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+350.^2]], "   # input scale\n", 
      " 1   ", FORTRANInputForm[300.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
      " 2   300.   # M2\n", 
      " 3   ", FORTRANInputForm[500.], "   # M3\n", 
      " 11   ", FORTRANInputForm[-770. + 300./OptionValue[tanbeta]], "   # At\n", 
      " 12   ", FORTRANInputForm[-770. + 300./OptionValue[tanbeta]], "   # Ab\n", 
      " 13   ", FORTRANInputForm[-770. + 300./OptionValue[tanbeta]], "   # Atau\n",
      " 23   300.   # mu\n", 
      " 26   ", FORTRANInputForm[OptionValue[MA]],"   # MA\n", 
      " 31   350.   # Msusy\n", 
      " 32   350.   # Msusy\n", 
      " 33   350.   # Msusy\n", 
      " 34   350.   # Msusy\n", 
      " 35   350.   # Msusy\n", 
      " 36   350.   # Msusy\n", 
      " 41   350.   # Msusy\n", 
      " 42   350.   # Msusy\n", 
      " 43   350.   # Msusy\n", 
      " 44   350.   # Msusy\n", 
      " 45   350.   # Msusy\n", 
      " 46   350.   # Msusy\n", 
      " 47   350.   # Msusy\n", 
      " 48   350.   # Msusy\n", 
      " 49   350.   # Msusy\n"], 
    "smallalpha", 
     WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   ", FORTRANInputForm[OptionValue[tanbeta]], "   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   ", FORTRANInputForm[Sqrt[OptionValue[MtOS]^2+800.^2]], "   # input scale\n", 
     " 1   ", FORTRANInputForm[500.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2)], "   # M1=5/3*sw^2/cw^2*M2=5/3*(Mz^2/Mw^2-1)*M2\n", 
     " 2   500.   # M2\n",
     " 3   500.   # M3\n", 
     " 11   ", FORTRANInputForm[-1200. + 2000./OptionValue[tanbeta]], "   # At\n",
     " 12   ", FORTRANInputForm[-1200. + 2000./OptionValue[tanbeta]], "   # Ab\n", 
     " 13   ", FORTRANInputForm[-1200. + 2000./OptionValue[tanbeta]], "   # Atau\n", 
     " 23   2000.   # mu\n", 
     " 26   ", FORTRANInputForm[OptionValue[MA]], "   # MA\n", 
     " 31   800.   # Msusy\n", 
     " 32   800.   # Msusy\n", 
     " 33   800.   # Msusy\n", 
     " 34   800.   # Msusy\n", 
     " 35   800.   # Msusy\n", 
     " 36   800.   # Msusy\n", 
     " 41   800.   # Msusy\n", 
     " 42   800.   # Msusy\n", 
     " 43   800.   # Msusy\n", 
     " 44   800.   # Msusy\n", 
     " 45   800.   # Msusy\n", 
     " 46   800.   # Msusy\n", 
     " 47   800.   # Msusy\n", 
     " 48   800.   # Msusy\n", 
     " 49   800.   # Msusy\n"],
    "p19MSSM1",(*  defined like in [arXiv:1109.3859 [hep-ph] 18 Sep 2011] on page 20 (27) (electroweake scale = Mz)*)
     WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   10.   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   ", FORTRANInputForm[OptionValue[Mz]], "   # input scale = ew scale = Mz \n", 
     " 1   ", FORTRANInputForm[OptionValue[M1]], "   # M1\n", 
     " 2   2500.0   # M2\n",
     " 3   ", FORTRANInputForm[OptionValue[M3]], "   # M3\n", 
     " 11   0.0   # At\n",
     " 12   0.0   # Ab\n", 
     " 13   0.0   # Atau\n", 
     " 23   2500.0   # mu\n", 
     " 26   2500.0   # MA\n", 
     " 31   ", FORTRANInputForm[OptionValue[M3]], "   # m~eL\n", 
     " 32   ", FORTRANInputForm[OptionValue[M3]], "   # m~muL\n", 
     " 33   2500.0   # m~tauL\n", 
     " 34   2500.0   # m~eR\n", 
     " 35   2500.0   # m~muR\n", 
     " 36   2500.0   # m~tauR\n", 
     " 41   ", FORTRANInputForm[OptionValue[M3]], "   # m~q1L\n", 
     " 42   ", FORTRANInputForm[OptionValue[M3]], "   # m~q2L\n", 
     " 43   2500.0   # m~q3L\n", 
     " 44   ", FORTRANInputForm[OptionValue[M3]], "   # m~uR\n", 
     " 45   ", FORTRANInputForm[OptionValue[M3]], "   # m~cR\n", 
     " 46   2500.0   # m~tR\n", 
     " 47   ", FORTRANInputForm[OptionValue[M3]], "   # m~dR\n", 
     " 48   ", FORTRANInputForm[OptionValue[M3]], "   # m~sR\n", 
     " 49   2500.0   # m~bR\n"],
    "p19MSSM2",(*  defined like in [arXiv:1109.3859 [hep-ph] 18 Sep 2011] on page 20 (28) (electroweake scale = Mz) *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   10.   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   ", FORTRANInputForm[OptionValue[Mz]], "   # input scale = electroweak scale = Mz \n", 
     " 1   ", FORTRANInputForm[OptionValue[M1]], "   # M1\n", 
     " 2   2500.0   # M2\n",
     " 3   2500.0   # M3\n", 
     " 11   0.0   # At\n",
     " 12   0.0   # Ab\n", 
     " 13   0.0   # Atau\n", 
     " 23   2500.0   # mu\n", 
     " 26   2500.0   # MA\n", 
     " 31   ", FORTRANInputForm[OptionValue[mslepton]], "   # m~eL\n", 
     " 32   ", FORTRANInputForm[OptionValue[mslepton]], "   # m~muL\n", 
     " 33   2500.0   # m~tauL\n", 
     " 34   ", FORTRANInputForm[OptionValue[mslepton]], "   # m~eR\n", 
     " 35   ", FORTRANInputForm[OptionValue[mslepton]], "   # m~muR\n", 
     " 36   2500.0   # m~tauR\n", 
     " 41   2500.0   # m~q1L\n", 
     " 42   2500.0   # m~q2L\n", 
     " 43   2500.0   # m~q3L\n", 
     " 44   2500.0   # m~uR\n", 
     " 45   2500.0   # m~cR\n", 
     " 46   2500.0   # m~tR\n", 
     " 47   2500.0   # m~dR\n", 
     " 48   2500.0   # m~sR\n", 
     " 49   2500.0   # m~bR\n"],
    "401479",(* (Bino-stop coannihilation) defined like in http://www.slac.stanford.edu/~aismail/snowmass/401479.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   29.1   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   1.45760400E+03   # input scale(MX) = mean(stmassparams)\n", 
     " 1   7.50910000E+02   # M_1(MX)\n", 
     " 2   -1.69685000E+03   # M_2(MX)\n",
     " 3   3.45419000E+03   # M_3(MX)\n", 
     " 11   -3.28602000E+03  # At(MX)\n",
     " 12   7.06530000E+02   # Ab(MX)\n", 
     " 13   -1.47756000E+03  # Atau(MX)\n", 
     " 23   2.74879000E+03   # mu(MX)\n", 
     " 26   3.52257000E+03   # mA(pole)\n", 
     " 31   2.30336000E+03   # meL(MX)\n", 
     " 32   2.30336000E+03   # mmuL(MX)\n", 
     " 33   9.14750000E+02   # mtauL(MX)\n", 
     " 34   2.10278000E+03   # meR(MX)\n", 
     " 35   2.10278000E+03   # mmuR(MX)\n", 
     " 36   1.47652000E+03   # mtauR(MX)\n", 
     " 41   3.71511000E+03   # mqL1(MX)\n", 
     " 42   3.71511000E+03   # mqL2(MX)\n", 
     " 43   8.99370000E+02   # mqL3(MX)\n", 
     " 44   3.98351000E+03   # muR(MX)\n", 
     " 45   3.98351000E+03   # mcR(MX)\n", 
     " 46   2.39244000E+03   # mtR(MX)\n", 
     " 47   2.79581000E+03   # mdR(MX)\n", 
     " 48   2.79581000E+03   # msR(MX)\n", 
     " 49   3.90657000E+03   # mbR(MX)\n"],
    "1046838",(* (Pure Higgsino) defined like in http://www.slac.stanford.edu/~aismail/snowmass/1046838.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   15.14   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   0.24648300E+04   # input scale(MX) = mean(stmassparams)\n", 
     " 1   -1.80245000E+03   # M_1(MX)\n", 
     " 2   3.26865000E+03   # M_2(MX)\n",
     " 3   2.35064000E+03   # M_3(MX)\n", 
     " 11   3.93959000E+03   # At(MX)\n",
     " 12   8.80380000E+02   # Ab(MX)\n", 
     " 13   -1.59228000E+03   # Atau(MX)\n", 
     " 23   1.03294000E+03   # mu(MX)\n", 
     " 26   1.76939000E+03   # mA(pole)\n", 
     " 31   3.11317000E+03   # meL(MX)\n", 
     " 32   3.11317000E+03   # mmuL(MX)\n", 
     " 33   2.29103000E+03   # mtauL(MX)\n", 
     " 34   3.55359000E+03   # meR(MX)\n", 
     " 35   3.55359000E+03   # mmuR(MX)\n", 
     " 36   1.65174000E+03   # mtauR(MX)\n", 
     " 41   2.97577000E+03   # mqL1(MX)\n", 
     " 42   2.97577000E+03   # mqL2(MX)\n", 
     " 43   2.56208000E+03   # mqL3(MX)\n", 
     " 44   3.63501000E+03   # muR(MX)\n", 
     " 45   3.63501000E+03   # mcR(MX)\n", 
     " 46   2.37387000E+03   # mtR(MX)\n", 
     " 47   1.60469000E+03   # mdR(MX)\n", 
     " 48   1.60469000E+03   # msR(MX)\n", 
     " 49   3.74657000E+03   # mbR(MX)\n"],
    "2342344",(* (Bino-squark coannihilation (Spoke 1 & Spoke 2)) defined like in http://www.slac.stanford.edu/~aismail/snowmass/2342344-slope.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   42.27   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   0.18446550E+04   # input scale(MX) ~ mean(stmassparams)\n", (* Note: this should be -1 ->\sqrt(\tilde{m_{t_1}}*\tilde{m_{t_2}}) because the input scale is changing with y and x but only softsusy gives proper scale with -1... *)
     " 1   ", FORTRANInputForm[-(861.42+16.*OptionValue[yBFS2013]+12.*OptionValue[xBFS2013])],"   # M_1(MX)\n", 
     " 2   ", FORTRANInputForm[2347.85+64.*OptionValue[yBFS2013]],"   # M_2(MX)\n",
     " 3   ", FORTRANInputForm[1175.60+64.*OptionValue[yBFS2013]+11*OptionValue[xBFS2013]],"   # M_3(MX)\n", 
     " 11   ", FORTRANInputForm[-(3869.01+64.*OptionValue[yBFS2013])],"   # At(MX)\n",
     " 12   ", FORTRANInputForm[595.75+64.*OptionValue[yBFS2013]],"   # Ab(MX)\n", 
     " 13   ", FORTRANInputForm[3643.08+64.*OptionValue[yBFS2013]],"   # Atau(MX)\n", 
     " 23   ", FORTRANInputForm[-(1888.05+64.*OptionValue[yBFS2013])],"   # mu(MX)\n", 
     " 26   ", FORTRANInputForm[3725.00+64.*OptionValue[yBFS2013]],"   # mA(pole)\n", 
     " 31   ", FORTRANInputForm[1358.97+64.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # meL(MX)\n", 
     " 32   ", FORTRANInputForm[1358.97+64.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # mmuL(MX)\n", 
     " 33   ", FORTRANInputForm[2135.21+64.*OptionValue[yBFS2013]],"   # mtauL(MX)\n", 
     " 34   ", FORTRANInputForm[2732.11+64.*OptionValue[yBFS2013]],"   # meR(MX)\n", 
     " 35   ", FORTRANInputForm[2732.11+64.*OptionValue[yBFS2013]],"   # mmuR(MX)\n", 
     " 36   ", FORTRANInputForm[2248.42+64.*OptionValue[yBFS2013]],"   # mtauR(MX)\n", 
     " 41   ", FORTRANInputForm[828.38+9.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # mqL1(MX)\n", 
     " 42   ", FORTRANInputForm[828.38+9.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # mqL2(MX)\n", 
     " 43   ", FORTRANInputForm[1072.25+64.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # mqL3(MX)\n", 
     " 44   ", FORTRANInputForm[805.97+9.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # muR(MX)\n", 
     " 45   ", FORTRANInputForm[805.97+9.*OptionValue[yBFS2013]+11.*OptionValue[xBFS2013]],"   # mcR(MX)\n", 
     " 46   ", FORTRANInputForm[3186.49+64.*OptionValue[yBFS2013]],"   # mtR(MX)\n", 
     " 47   ", FORTRANInputForm[1903.76+64.*OptionValue[yBFS2013]],"   # mdR(MX)\n", 
     " 48   ", FORTRANInputForm[1903.76+64.*OptionValue[yBFS2013]],"   # msR(MX)\n", 
     " 49   ", FORTRANInputForm[3064.69+64.*OptionValue[yBFS2013]],"   # mbR(MX)\n"],
    "2387564",(* (A funnel) defined like in http://www.slac.stanford.edu/~aismail/snowmass/2387564.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   11.73   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   0.23848210E+04   # input scale(MX) = mean(stmassparams)\n", 
     " 1   1.01969000E+03   # M_1(MX)\n", 
     " 2   1.54922000E+03   # M_2(MX)\n",
     " 3   2.51454000E+03   # M_3(MX)\n", 
     " 11   3.94180000E+03   # At(MX)\n",
     " 12   -9.32660000E+02   # Ab(MX)\n", 
     " 13   2.23973000E+03   # Atau(MX)\n", 
     " 23   -1.77575000E+03   # mu(MX)\n", 
     " 26   2.04273000E+03   # mA(pole)\n", 
     " 31   2.93029000E+03   # meL(MX)\n", 
     " 32   2.93029000E+03   # mmuL(MX)\n", 
     " 33   2.80032000E+03   # mtauL(MX)\n", 
     " 34   1.47379000E+03   # meR(MX)\n", 
     " 35   1.47379000E+03   # mmuR(MX)\n", 
     " 36   2.55030000E+03   # mtauR(MX)\n", 
     " 41   1.31150000E+03   # mqL1(MX)\n", 
     " 42   1.31150000E+03   # mqL2(MX)\n", 
     " 43   2.60547000E+03   # mqL3(MX)\n", 
     " 44   3.02443000E+03   # muR(MX)\n", 
     " 45   3.02443000E+03   # mcR(MX)\n", 
     " 46   2.18727000E+03   # mtR(MX)\n", 
     " 47   3.40446000E+03   # mdR(MX)\n", 
     " 48   3.40446000E+03   # msR(MX)\n", 
     " 49   3.96410000E+03   # mbR(MX)\n"],
    "2750334",(* (Well-tempered neutralino) defined like in http://www.slac.stanford.edu/~aismail/snowmass/2750334.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   52.99   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   2.36175100E+03   # input scale(MX) = mean(stmassparams)\n", 
     " 1   -1.69300000E+02   # M_1(MX)\n", 
     " 2   -1.63609000E+03   # M_2(MX)\n",
     " 3   1.75526000E+03   # M_3(MX)\n", 
     " 11   3.79906000E+03   # At(MX)\n",
     " 12   -3.25554000E+03   # Ab(MX)\n", 
     " 13   -2.90852000E+03   # Atau(MX)\n", 
     " 23   -1.92450000E+02   # mu(MX)\n", 
     " 26   1.39988000E+03   # mA(pole)\n", 
     " 31   3.28068000E+03   # meL(MX)\n", 
     " 32   3.28068000E+03   # mmuL(MX)\n", 
     " 33   1.59063000E+03   # mtauL(MX)\n", 
     " 34   7.45880000E+02   # meR(MX)\n", 
     " 35   7.45880000E+02   # mmuR(MX)\n", 
     " 36   2.17791000E+03   # mtauR(MX)\n", 
     " 41   2.51894000E+03   # mqL1(MX)\n", 
     " 42   2.51894000E+03   # mqL2(MX)\n", 
     " 43   2.47243000E+03   # mqL3(MX)\n", 
     " 44   1.35954000E+03   # muR(MX)\n", 
     " 45   1.35954000E+03   # mcR(MX)\n", 
     " 46   2.25927000E+03   # mtR(MX)\n", 
     " 47   1.15863000E+03   # mdR(MX)\n", 
     " 48   1.15863000E+03   # msR(MX)\n", 
     " 49   2.33404000E+03   # mbR(MX)\n"],
    "374345",(* (Super-WIMP) defined like in http://www.slac.stanford.edu/~aismail/snowmass/374345.txt *)
      WriteLineFunction[InputFileString, 
     "Block MINPAR\n", 
     " 3   48.22   # tan(beta)\n", 
     "Block EXTPAR\n", 
     " 0   0.19393880E+04   # input scale(MX) = mean(stmassparams)\n", 
     " 1   -3.21050000E+03   # M_1(MX)\n", 
     " 2   2.87559000E+03   # M_2(MX)\n",
     " 3   2.05302000E+03   # M_3(MX)\n", 
     " 11   3.43314000E+03   # At(MX)\n",
     " 12   -2.83087000E+03   # Ab(MX)\n", 
     " 13   -3.92010000E+03   # Atau(MX)\n", 
     " 23   1.19322000E+03   # mu(MX)\n", 
     " 26   2.08325000E+03   # mA(pole)\n", 
     " 31   5.74640000E+02   # meL(MX) \n", 
     " 32   5.74640000E+02   # mmuL(MX)\n", 
     " 33   6.35890000E+02   # mtauL(MX)\n", 
     " 34   2.48188000E+03   # meR(MX)\n", 
     " 35   2.48188000E+03   # mmuR(MX)\n", 
     " 36   3.56968000E+03   # mtauR(MX)\n", 
     " 41   3.08437000E+03   # mqL1(MX)\n", 
     " 42   3.08437000E+03   # mqL2(MX)\n", 
     " 43   3.63741000E+03   # mqL3(MX)\n", 
     " 44   2.04632000E+03   # muR(MX)\n", 
     " 45   2.04632000E+03   # mcR(MX)\n", 
     " 46   1.03271000E+03   # mtR(MX)\n", 
     " 47   7.02920000E+02   # mdR(MX)\n", 
     " 48   7.02920000E+02   # msR(MX)\n", 
     " 49   2.13942000E+03   # mbR(MX)\n"]];

    If[OptionValue[InputFilePath]===String,
       Return[InputFileString];,
       Close[InputFileString];
    ];

 
 ];


(* Converting Mathematica input into FORTRAN readable output*)
LSHInverseProcessValue[x_, String] := ToString[x];
LSHInverseProcessValue[x_, Integer] := ToString[x];
LSHInverseProcessValue[x_, Real] := If[NumberQ[x], FORTRANInputForm[x], ToString[x]];

LSHInverseProcessValue[x_, Automatic]:= Which[Head[x]===Integer,
                                               ToString[x],
                                              NumberQ[x],
                                               FORTRANInputForm[x],
                                              Head[x]===String,
                                               ToString[x]];


DefaultInputList={"MODSEL" -> {
   BC -> None, 
   BI -> {EV -> None, EC -> None}, 
   11 -> {EV -> 1, EC -> " One value of the renormalisation scale"}, 
   12 -> {EV -> Global`MtOSValue, EC -> " renormalisation scale"}, 
   1 -> {EV -> 1, EC -> " msugra"}}, 
 "SPHENOINPUT" -> {
   BC -> None, 
   BI -> {EV -> None, EC -> None}, 
   11 -> {EV -> 0, EC -> " don't calculate branching ratios"}, 
   21 -> {EV -> 0, EC -> " don't calculate cross-sections"}}, 
 "SMINPUTS" -> {
   BC -> None, 
   BI -> {EV -> None, EC -> None}, 
   1 -> {EV -> Global`invalphaValue, EC -> " alpha^(-1) SM MSbar(Mz)"}, 
   2 -> {EV -> Global`GFValue, EC -> " G_Fermi"}, 
   3 -> {EV -> Global`asmzValue, EC -> " alpha_s(Mz) SM MSbar"}, 
   4 -> {EV -> Global`MzValue, EC -> " Mz(pole)"},
    5 -> {EV -> Global`MbMSbarValue, EC -> " mb(mb) SM MSbar"}, 
   6 -> {EV -> Global`MtOSValue, EC -> " Mt SM pole"}, 
   7 -> {EV -> Global`MtauValue, EC -> " mtau pole"}}, 
 "MINPAR" -> {
   BC -> None, 
   BI -> {EV -> None, EC -> None}, 
   1 -> {EV -> 1.00000000*10^(2), EC -> " m_0"}, 
   2 -> {EV -> 2.50000000*10^(2), EC -> " m_1/2"}, 
   3 -> {EV -> 1.00000000*10^(1), EC -> " tan(beta)"}, 
   4 -> {EV -> 1.00000000, EC -> "sign(mu)"}, 
   5 -> {EV -> -1.00000000*10^(2), EC -> " A0"}}};

(* Function wich writes LesHouches-Files with internal Mathematica Representations as input*)
Options[WriteLesHouchesFile] = {InputFilePath -> Global`InputFilePathValue, 
                                InputList -> DefaultInputList,
                                AppendToFile->False 
                                (*,Debug -> False*)};

WriteLesHouchesFile[opt : OptionsPattern[]] := Module[{OutputString, StringAppend, Status, CurrentBlockName, CurrentEntry, CurrentBI, CurrentBC},
   
(*    If[OptionValue[Debug], Status = Print]; *)
   
   OutputString = "# SLHA input file generated by SLAM.m\n";
   StringAppend[x_] := (OutputString = OutputString <> x);
   
   
   (CurrentBlockName = #[[1]];
(*       Status["Writing new block name: ", CurrentBlockName = #[[1]]]; *)
      StringAppend["Block " <> CurrentBlockName <> " "];
      
      (CurrentBI = EV /. ((BI /. #)/.BI->{});
         If[(CurrentBI =!= None)&& (CurrentBI =!= EV),
(*           Status["Writing block info: ", CurrentBI]; *)
          StringAppend[CurrentBI <> " "];
          ];
         
         CurrentBC = BC /. #;
         If[(CurrentBC =!= None)&&(CurrentBC=!= BC),
(*           Status["Writing block comment: ", CurrentBC]; *)
          StringAppend["# " <> CurrentBC <> " \n"];,
          StringAppend["# \n"];
          ];
         
(*          Status["Writing entries..."]; *)
         
         (
            Which[(Length[CurrentEntry = #[[1]]] === 2),
             (*2 fold entries*)
             
(*              Status["Writing 2-fold entry: ", #[[1]]]; *)
             StringAppend[" "<>ToString[#[[1, 1]]] <> " " <> ToString[#[[1, 2]]] <> "   "];,
             (Length[CurrentEntry = #[[1]]] === 0) && (#[[1]] =!= None),
             (*1 fold entries*)
             
(*              Status["Writing 1-fold entry: ", #[[1]]]; *)
             StringAppend[" "<>ToString[#[[1]]] <> "   "];,
             (#[[1]] === None),
             (*None entries*)
             
(*              Status["Writing 0-fold entry: ", #[[1]]]; *)
             StringAppend["         "];
             ];
            
            ((*Status["Writing entry value:", EV /. #];*)
               
               StringAppend[LSHInverseProcessValue[EV /. #, Automatic] <> "   "];
               (*Status["Writing entry comment:", EC /. #];*)
               If[((EC /. #) =!= None) && ((EC /. #)=!= EC),
                StringAppend["# " <> ToString[EC /. #] <> " \n"];,
                StringAppend["# \n"];
                ];) &@(#[[2]]);
            ) & /@ Complement[#, FilterRules[#, {"BC", "BI"}]]
         ) &@(Sort[#[[2]]]);
      ) & /@ OptionValue[InputList];
   
   Which[OptionValue[InputFilePath] === String,
    Return[OutputString];,
    Head[OptionValue[InputFilePath]] === String,
    If[OptionValue[AppendToFile]&&(FileType[OptionValue[InputFilePath]]===File),
       Export[OptionValue[InputFilePath], Import[OptionValue[InputFilePath]]<>"\n"<>OutputString, "Text"];,
       Export[OptionValue[InputFilePath], OutputString, "Text"];
    ];
    ];
   
 ];



Options[GenerateInputSQLChecklist] = Options[WriteSpectrumGeneratorInput];

(* StableNumber is used because there can be rounding errors making SLAM not able to find a spectrum in data base although it has been saved  but including the rounding error*)
StableNumber= FORTRANToMathematicaNumber[FORTRANInputForm[#]]&;

GenerateInputSQLChecklist[opt : OptionsPattern[]] := {
 Sequence @@ If[NumberQ[OptionValue[muRen]] && (OptionValue[SpectrumGenerator]=!="suspect"),(* suspect cannot deal with renormalisation scale *)
      {{"I_MODSEL_11", 1},
       {"I_MODSEL_12", StableNumber @ (1.0 * OptionValue[muRen])}},{}],
      {"I_MODSEL_1", Switch[OptionValue[Model],
                        "msugra", 1,
                        "mgmsb", 2,
                        "mamsb", 3,
                        "nomix", 0,
                        "gluophob", 0,
                        "smallalpha", 0,
                        "mhmax", 0,
                        "mhmaxup", 0,
                        "mhmod+", 0,
                        "mhmod-", 0,
                        "lightstop", 0,
                        "lightstau", 0,
                        "lightstaudeltatau", 0,
                        "tauphobic", 0,
                        "lowmh", 0,
                        "p19MSSM1", 0,
                        "p19MSSM2", 0,
                        "401479", 0,
                        "1046838", 0,
                        "2342344", 0,
                        "2387564", 0,
                        "2750334", 0,
                        "374345", 0,
                        _, Message[SLAM::UnknownModel,OptionValue[Model]];Abort[];]},
 Sequence @@ Switch[OptionValue[SpectrumGenerator],
   "softsusy", 
      {{"I_SOFTSUSY_5", 1}},
   "spheno",
     {{"I_SPHENOINPUT_11", 0}, {"I_SPHENOINPUT_21", 0}},
   "suspect",
     {{"I_SU_ALGO_2",21}},
    _,
   Message[SLAM::UnknownSpecGen];Abort[];],
     {"I_SMINPUTS_1",StableNumber @ (1.0 * OptionValue[invalpha])},
     {"I_SMINPUTS_2",StableNumber @ (1.0 * OptionValue[GF])},
     {"I_SMINPUTS_3",StableNumber @ (1.0 * OptionValue[asmz])},
     {"I_SMINPUTS_4",StableNumber @ (1.0 * OptionValue[Mz])},
     {"I_SMINPUTS_5",StableNumber @ (1.0 * OptionValue[mb])},
     {"I_SMINPUTS_6",StableNumber @ (1.0 * OptionValue[MtOS])},
     {"I_SMINPUTS_7",StableNumber @ (1.0 * OptionValue[Mtau])},
  Sequence @@ Switch[OptionValue[Model],
   "msugra", {
     {"I_MINPAR_1",StableNumber @ (1.0 * OptionValue[m0])},
     {"I_MINPAR_2",StableNumber @ (1.0 * OptionValue[m12])},
     {"I_MINPAR_3",StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_MINPAR_4",StableNumber @ (1.0 * OptionValue[signmu])},
     {"I_MINPAR_5",StableNumber @ (1.0 * OptionValue[A0])}},
   "mgmsb", {
     {"I_MINPAR_1",StableNumber @ (1.0 * OptionValue[lambda])},
     {"I_MINPAR_2",StableNumber @ (1.0 * OptionValue[mmess])},
     {"I_MINPAR_3",StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_MINPAR_4",StableNumber @ (1.0 * OptionValue[signmu])},
     {"I_MINPAR_5",StableNumber @ (1.0 * OptionValue[nmess])}},
   "mamsb", {
     {"I_MINPAR_1",StableNumber @ (1.0 * OptionValue[m0])},
     {"I_MINPAR_2",StableNumber @ (1.0 * OptionValue[m32])},
     {"I_MINPAR_3",StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_MINPAR_4",StableNumber @ (1.0 * OptionValue[signmu])}},
   "mhmax", {
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 800.},
     {"I_EXTPAR_11", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 200.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 1000.},
     {"I_EXTPAR_32", 1000.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 1000.},
     {"I_EXTPAR_35", 1000.},
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1000.},
     {"I_EXTPAR_42", 1000.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1000.},
     {"I_EXTPAR_45", 1000.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1000.},
     {"I_EXTPAR_48", 1000.},
     {"I_EXTPAR_49", 1000.}
     },
   "mhmaxup",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (Sqrt[6]*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 200.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1000.}
     }, 
   "mhmod+",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (1.6*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (1.6*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (1.6*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 200.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1000.}
     }, 
   "mhmod-",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (-2.2*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (-2.2*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (-2.2*1000. + 200./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 200.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1000.}
     }, 
  "lightstop",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},(* Note: Msusy=500. but this is too low for input scale (->tachyonic particles ) thats why we put 1000. In agreement with OS masses for st1 when comparing softsusy and feynhiggs  *)
     {"I_EXTPAR_1", StableNumber @ (350.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 350.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (2.2*500. + 350./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (2.2*500. + 350./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (2.2*500. + 350./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 350.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 500.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 500.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 500.}
     },
  "lightstau",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (1.7*1000. + 500./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (1.7*1000. + 500./OptionValue[tanbeta])},
     {"I_EXTPAR_13", 0.},
     {"I_EXTPAR_23", 500.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 245.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 245.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1000.}
     }, 
  "lightstaudeltatau",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1000.^2]},
     {"I_EXTPAR_1", StableNumber @ (400.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 400.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (1.7*1000. + 450./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (1.7*1000. + 450./OptionValue[tanbeta])},
     {"I_EXTPAR_13", 0.},
     {"I_EXTPAR_23", 450.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 250.},
     {"I_EXTPAR_34", 500. },
     {"I_EXTPAR_35", 500. },
     {"I_EXTPAR_36", 250.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1000.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1000.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1000.}
     }, 
  "tauphobic",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1500.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (2.9*1500. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (2.9*1500. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (2.9*1500. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 2000.},
     {"I_EXTPAR_26", StableNumber @ (1.0*OptionValue[MA])},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 500.},
     {"I_EXTPAR_34", 500.},
     {"I_EXTPAR_35", 500.},
     {"I_EXTPAR_36", 500.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1500.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1500.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1500.}
     },
  "lowmh",{
     {"I_MINPAR_3", StableNumber @ (1.0 * OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+1500.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1500.},
     {"I_EXTPAR_11", StableNumber @ (2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (2.9*1500. + 1.0*OptionValue[mu]/OptionValue[tanbeta])},
     {"I_EXTPAR_23", StableNumber @ (1.0*OptionValue[mu])},
     {"I_EXTPAR_26", 110.},
     {"I_EXTPAR_31", 500.},
     {"I_EXTPAR_32", 500.},
     {"I_EXTPAR_33", 1000.},
     {"I_EXTPAR_34", 500.},
     {"I_EXTPAR_35", 500.},
     {"I_EXTPAR_36", 1000.},
     {"I_EXTPAR_41", 1500.},
     {"I_EXTPAR_42", 1500.},
     {"I_EXTPAR_43", 1500.},
     {"I_EXTPAR_44", 1500.},
     {"I_EXTPAR_45", 1500.},
     {"I_EXTPAR_46", 1500.},
     {"I_EXTPAR_47", 1500.},
     {"I_EXTPAR_48", 1500.},
     {"I_EXTPAR_49", 1500.}
     }, 
 "nomix", {
     {"I_MINPAR_3",StableNumber @  (1.0 *OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+2000.^2]},
     {"I_EXTPAR_1", StableNumber @ (200.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 200.},
     {"I_EXTPAR_3", 1600.},
     {"I_EXTPAR_11", StableNumber @ (200./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (200./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (200./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 200.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 2000.},
     {"I_EXTPAR_32", 2000.},
     {"I_EXTPAR_33", 2000.},
     {"I_EXTPAR_34", 2000.},
     {"I_EXTPAR_35", 2000.},
     {"I_EXTPAR_36", 2000.},
     {"I_EXTPAR_41", 2000.},
     {"I_EXTPAR_42", 2000.},
     {"I_EXTPAR_43", 2000.},
     {"I_EXTPAR_44", 2000.},
     {"I_EXTPAR_45", 2000.},
     {"I_EXTPAR_46", 2000.},
     {"I_EXTPAR_47", 2000.},
     {"I_EXTPAR_48", 2000.},
     {"I_EXTPAR_49", 2000.}
     },
"gluophob", {
     {"I_MINPAR_3", StableNumber @ (1.0 *OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+350.^2]},
     {"I_EXTPAR_1", StableNumber @ (300.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 300.},
     {"I_EXTPAR_3", 500.},
     {"I_EXTPAR_11", StableNumber @ (-770. + 300./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (-770. + 300./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (-770. + 300./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 300},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 350.},
     {"I_EXTPAR_32", 350.},
     {"I_EXTPAR_33", 350.},
     {"I_EXTPAR_34", 350.},
     {"I_EXTPAR_35", 350.},
     {"I_EXTPAR_36", 350.},
     {"I_EXTPAR_41", 350.},
     {"I_EXTPAR_42", 350.},
     {"I_EXTPAR_43", 350.},
     {"I_EXTPAR_44", 350.},
     {"I_EXTPAR_45", 350.},
     {"I_EXTPAR_46", 350.},
     {"I_EXTPAR_47", 350.},
     {"I_EXTPAR_48", 350.},
     {"I_EXTPAR_49", 350.}
     },
 "smallalpha", {
     {"I_MINPAR_3", StableNumber @ (1.0 *OptionValue[tanbeta])},
     {"I_EXTPAR_0", StableNumber @ Sqrt[OptionValue[MtOS]^2+800.^2]},
     {"I_EXTPAR_1", StableNumber @ (500.*5/3*(-1+OptionValue[Mz]^2/OptionValue[Mw]^2))},
     {"I_EXTPAR_2", 500.},
     {"I_EXTPAR_3", 500.},
     {"I_EXTPAR_11", StableNumber @ (-1200. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_12", StableNumber @ (-1200. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_13", StableNumber @ (-1200. + 2000./OptionValue[tanbeta])},
     {"I_EXTPAR_23", 2000.},
     {"I_EXTPAR_26", StableNumber @ (1.0 *OptionValue[MA])},
     {"I_EXTPAR_31", 800.},
     {"I_EXTPAR_32", 800.},
     {"I_EXTPAR_33", 800.},
     {"I_EXTPAR_34", 800.},
     {"I_EXTPAR_35", 800.},
     {"I_EXTPAR_36", 800.},
     {"I_EXTPAR_41", 800.},
     {"I_EXTPAR_42", 800.},
     {"I_EXTPAR_43", 800.},
     {"I_EXTPAR_44", 800.},
     {"I_EXTPAR_45", 800.},
     {"I_EXTPAR_46", 800.},
     {"I_EXTPAR_47", 800.},
     {"I_EXTPAR_48", 800.},
     {"I_EXTPAR_49", 800.}
     },
 "p19MSSM1",{
     {"I_MINPAR_3", 10.},
     {"I_EXTPAR_0", StableNumber @ (1.0 *OptionValue[Mz])},
     {"I_EXTPAR_1", StableNumber @ (1.0 *OptionValue[M1])},
     {"I_EXTPAR_2", 2500.0},
     {"I_EXTPAR_3", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_11", 0.0},
     {"I_EXTPAR_12", 0.0},
     {"I_EXTPAR_13", 0.0},
     {"I_EXTPAR_23", 2500.0},
     {"I_EXTPAR_26", 2500.0},
     {"I_EXTPAR_31", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_32", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_33", 2500.0},
     {"I_EXTPAR_34", 2500.0},
     {"I_EXTPAR_35", 2500.0},
     {"I_EXTPAR_36", 2500.0},
     {"I_EXTPAR_41", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_42", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_43", 2500.0},
     {"I_EXTPAR_44", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_45", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_46", 2500.0},
     {"I_EXTPAR_47", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_48", StableNumber @ (1.0 *OptionValue[M3])},
     {"I_EXTPAR_49", 2500.0}
     },
 "p19MSSM2",{
     {"I_MINPAR_3", 10.},
     {"I_EXTPAR_0", StableNumber @ (1.0 *OptionValue[Mz])},
     {"I_EXTPAR_1", StableNumber @ (1.0 *OptionValue[M1])},
     {"I_EXTPAR_2", 2500.0},
     {"I_EXTPAR_3", 2500.0},
     {"I_EXTPAR_11", 0.0},
     {"I_EXTPAR_12", 0.0},
     {"I_EXTPAR_13", 0.0},
     {"I_EXTPAR_23", 2500.0},
     {"I_EXTPAR_26", 2500.0},
     {"I_EXTPAR_31", StableNumber @ (1.0 *OptionValue[mslepton])},
     {"I_EXTPAR_32", StableNumber @ (1.0 *OptionValue[mslepton])},
     {"I_EXTPAR_33", 2500.0},
     {"I_EXTPAR_34", StableNumber @ (1.0 *OptionValue[mslepton])},
     {"I_EXTPAR_35", StableNumber @ (1.0 *OptionValue[mslepton])},
     {"I_EXTPAR_36", 2500.0},
     {"I_EXTPAR_41", 2500.0},
     {"I_EXTPAR_42", 2500.0},
     {"I_EXTPAR_43", 2500.0},
     {"I_EXTPAR_44", 2500.0},
     {"I_EXTPAR_45", 2500.0},
     {"I_EXTPAR_46", 2500.0},
     {"I_EXTPAR_47", 2500.0},
     {"I_EXTPAR_48", 2500.0},
     {"I_EXTPAR_49", 2500.0}
     },
 "401479",{
     {"I_MINPAR_3", StableNumber @ 29.1},
     {"I_EXTPAR_0", StableNumber @ 1457.60400},
     {"I_EXTPAR_1", StableNumber @ 750.910000},
     {"I_EXTPAR_2", StableNumber @ -1696.85000},
     {"I_EXTPAR_3", StableNumber @ 3454.19000},
     {"I_EXTPAR_11", StableNumber @ -3286.02000},
     {"I_EXTPAR_12", StableNumber @ 706.530000},
     {"I_EXTPAR_13", StableNumber @ -1477.56000},
     {"I_EXTPAR_23", StableNumber @ 2748.79000},
     {"I_EXTPAR_26", StableNumber @ 3522.57000},
     {"I_EXTPAR_31", StableNumber @ 2303.36000},
     {"I_EXTPAR_32", StableNumber @ 2303.36000},
     {"I_EXTPAR_33", StableNumber @ 914.750000},
     {"I_EXTPAR_34", StableNumber @ 2102.78000},
     {"I_EXTPAR_35", StableNumber @ 2102.78000},
     {"I_EXTPAR_36", StableNumber @ 1476.52000},
     {"I_EXTPAR_41", StableNumber @ 3715.11000},
     {"I_EXTPAR_42", StableNumber @ 3715.11000},
     {"I_EXTPAR_43", StableNumber @ 899.370000},
     {"I_EXTPAR_44", StableNumber @ 3983.51000},
     {"I_EXTPAR_45", StableNumber @ 3983.51000},
     {"I_EXTPAR_46", StableNumber @ 2392.44000},
     {"I_EXTPAR_47", StableNumber @ 2795.81000},
     {"I_EXTPAR_48", StableNumber @ 2795.81000},
     {"I_EXTPAR_49", StableNumber @ 3906.57000}
     },
 "1046838",{
     {"I_MINPAR_3", StableNumber @ 15.14},
     {"I_EXTPAR_0", StableNumber @ 2464.83000},
     {"I_EXTPAR_1", StableNumber @ -1802.45000},
     {"I_EXTPAR_2", StableNumber @ 3268.65000},
     {"I_EXTPAR_3", StableNumber @ 2350.64000},
     {"I_EXTPAR_11", StableNumber @ 3939.59000},
     {"I_EXTPAR_12", StableNumber @ 880.380000},
     {"I_EXTPAR_13", StableNumber @ -1592.28000},
     {"I_EXTPAR_23", StableNumber @ 1032.94000},
     {"I_EXTPAR_26", StableNumber @ 1769.39000},
     {"I_EXTPAR_31", StableNumber @ 3113.17000},
     {"I_EXTPAR_32", StableNumber @ 3113.17000},
     {"I_EXTPAR_33", StableNumber @ 2291.03000},
     {"I_EXTPAR_34", StableNumber @ 3553.59000},
     {"I_EXTPAR_35", StableNumber @ 3553.59000},
     {"I_EXTPAR_36", StableNumber @ 1651.74000},
     {"I_EXTPAR_41", StableNumber @ 2975.77000},
     {"I_EXTPAR_42", StableNumber @ 2975.77000},
     {"I_EXTPAR_43", StableNumber @ 2562.08000},
     {"I_EXTPAR_44", StableNumber @ 3635.01000},
     {"I_EXTPAR_45", StableNumber @ 3635.01000},
     {"I_EXTPAR_46", StableNumber @ 2373.87000},
     {"I_EXTPAR_47", StableNumber @ 1604.69000},
     {"I_EXTPAR_48", StableNumber @ 1604.69000},
     {"I_EXTPAR_49", StableNumber @ 3746.57000}
     },
 "2342344",{
     {"I_MINPAR_3", StableNumber @ 42.27},
     {"I_EXTPAR_0", StableNumber @ 1844.65500},(* Note: this should be -1 ->\sqrt(\tilde{m_{t_1}}*\tilde{m_{t_2}}) because the input scale is changing with y and x but only softsusy gives proper scale with -1... *)
     {"I_EXTPAR_1", StableNumber @ (-(861.42+16*OptionValue[yBFS2013]+12*OptionValue[xBFS2013]))},
     {"I_EXTPAR_2", StableNumber @ (2347.85+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_3", StableNumber @ (1175.60+64*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_11", StableNumber @ (-(3869.01+64*OptionValue[yBFS2013]))},
     {"I_EXTPAR_12", StableNumber @ (595.75+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_13", StableNumber @ (3643.08+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_23", StableNumber @ (-(1888.05+64*OptionValue[yBFS2013]))},
     {"I_EXTPAR_26", StableNumber @ (3725.00+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_31", StableNumber @ (1358.97+64*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_32", StableNumber @ (1358.97+64*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_33", StableNumber @ (2135.21+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_34", StableNumber @ (2732.11+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_35", StableNumber @ (2732.11+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_36", StableNumber @ (2248.42+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_41", StableNumber @ (828.38+9*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_42", StableNumber @ (828.38+9*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_43", StableNumber @ (1072.25+64*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_44", StableNumber @ (805.97+9*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_45", StableNumber @ (805.97+9*OptionValue[yBFS2013]+11*OptionValue[xBFS2013])},
     {"I_EXTPAR_46", StableNumber @ (3186.49+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_47", StableNumber @ (1903.76+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_48", StableNumber @ (1903.76+64*OptionValue[yBFS2013])},
     {"I_EXTPAR_49", StableNumber @ (3064.69+64*OptionValue[yBFS2013])}
     },
 "2387564",{
     {"I_MINPAR_3", StableNumber @ 11.73},
     {"I_EXTPAR_0", StableNumber @ 2384.82100},
     {"I_EXTPAR_1", StableNumber @ 1019.69000},
     {"I_EXTPAR_2", StableNumber @ 1549.22000},
     {"I_EXTPAR_3", StableNumber @ 2514.54000},
     {"I_EXTPAR_11", StableNumber @ 3941.80000},
     {"I_EXTPAR_12", StableNumber @ -932.660000},
     {"I_EXTPAR_13", StableNumber @ 2239.73000},
     {"I_EXTPAR_23", StableNumber @ -1775.75000},
     {"I_EXTPAR_26", StableNumber @ 2042.73000},
     {"I_EXTPAR_31", StableNumber @ 2930.29000},
     {"I_EXTPAR_32", StableNumber @ 2930.29000},
     {"I_EXTPAR_33", StableNumber @ 2800.32000},
     {"I_EXTPAR_34", StableNumber @ 1473.79000},
     {"I_EXTPAR_35", StableNumber @ 1473.79000},
     {"I_EXTPAR_36", StableNumber @ 2550.30000},
     {"I_EXTPAR_41", StableNumber @ 1311.50000},
     {"I_EXTPAR_42", StableNumber @ 1311.50000},
     {"I_EXTPAR_43", StableNumber @ 2605.47000},
     {"I_EXTPAR_44", StableNumber @ 3024.43000},
     {"I_EXTPAR_45", StableNumber @ 3024.43000},
     {"I_EXTPAR_46", StableNumber @ 2187.27000},
     {"I_EXTPAR_47", StableNumber @ 3404.46000},
     {"I_EXTPAR_48", StableNumber @ 3404.46000},
     {"I_EXTPAR_49", StableNumber @ 3964.10000}
     },
 "2750334",{
     {"I_MINPAR_3", StableNumber @ 52.99},
     {"I_EXTPAR_0", StableNumber @ 2361.75100},
     {"I_EXTPAR_1", StableNumber @ -169.300000},
     {"I_EXTPAR_2", StableNumber @ -1636.09000},
     {"I_EXTPAR_3", StableNumber @ 1755.26000},
     {"I_EXTPAR_11", StableNumber @  3799.06000},
     {"I_EXTPAR_12", StableNumber @ -3255.54000},
     {"I_EXTPAR_13", StableNumber @ -2908.52000},
     {"I_EXTPAR_23", StableNumber @ -192.450000},
     {"I_EXTPAR_26", StableNumber @ 1399.88000},
     {"I_EXTPAR_31", StableNumber @ 3280.68000},
     {"I_EXTPAR_32", StableNumber @ 3280.68000},
     {"I_EXTPAR_33", StableNumber @ 1590.63000},
     {"I_EXTPAR_34", StableNumber @ 745.880000},
     {"I_EXTPAR_35", StableNumber @ 745.880000},
     {"I_EXTPAR_36", StableNumber @ 2177.91000},
     {"I_EXTPAR_41", StableNumber @ 2518.94000},
     {"I_EXTPAR_42", StableNumber @ 2518.94000},
     {"I_EXTPAR_43", StableNumber @ 2472.43000},
     {"I_EXTPAR_44", StableNumber @ 1359.54000},
     {"I_EXTPAR_45", StableNumber @ 1359.54000},
     {"I_EXTPAR_46", StableNumber @ 2259.27000},
     {"I_EXTPAR_47", StableNumber @ 1158.63000},
     {"I_EXTPAR_48", StableNumber @ 1158.63000},
     {"I_EXTPAR_49", StableNumber @ 2334.04000}
     },
 "374345",{
     {"I_MINPAR_3", StableNumber @ 48.22},
     {"I_EXTPAR_0", StableNumber @ 1939.38800},
     {"I_EXTPAR_1", StableNumber @ -3210.50000},
     {"I_EXTPAR_2", StableNumber @ 2875.59000},
     {"I_EXTPAR_3", StableNumber @ 2053.02000},
     {"I_EXTPAR_11", StableNumber @  3433.14000},
     {"I_EXTPAR_12", StableNumber @ -2830.87000},
     {"I_EXTPAR_13", StableNumber @ -3920.10000},
     {"I_EXTPAR_23", StableNumber @ 1193.22000},
     {"I_EXTPAR_26", StableNumber @ 2083.25000},
     {"I_EXTPAR_31", StableNumber @ 574.640000},
     {"I_EXTPAR_32", StableNumber @ 574.640000},
     {"I_EXTPAR_33", StableNumber @ 635.890000},
     {"I_EXTPAR_34", StableNumber @ 2481.88000},
     {"I_EXTPAR_35", StableNumber @ 2481.88000},
     {"I_EXTPAR_36", StableNumber @ 3569.68000},
     {"I_EXTPAR_41", StableNumber @ 3084.37000},
     {"I_EXTPAR_42", StableNumber @ 3084.37000},
     {"I_EXTPAR_43", StableNumber @ 3637.41000},
     {"I_EXTPAR_44", StableNumber @ 2046.32000},
     {"I_EXTPAR_45", StableNumber @ 2046.32000},
     {"I_EXTPAR_46", StableNumber @ 1032.71000},
     {"I_EXTPAR_47", StableNumber @ 702.920000},
     {"I_EXTPAR_48", StableNumber @ 702.920000},
     {"I_EXTPAR_49", StableNumber @ 2139.42000}
     }]};



(* Replacing a Value of a Rule by a new Value.
Example:ReplaceRule[{a->1,b->2}, a->3]   ->  {a->3,b->2}  *)
ReplaceRule[ListOfRules_List, NewRule_] := (ReplacePart[ListOfRules, Position[ListOfRules, (NewRule[[1]] -> _)] -> NewRule]);

(* Checks the Folder for already existing files who have BaseFilename in their name and generates a unique filename using numbering extensions *)
GetUniqueFileName[FolderName_String, BaseFilename_String] := Module[{FileCopyList},
   SetDirectory[FolderName];
   FileCopyList = (ToExpression[StringReplace[#, {RegularExpression[BaseFilename <> "(\\d+)$"] :> "$1", RegularExpression[".*"] -> ""}]] /. Null -> 0 & /@ (FileNames[{# <> "*"}] &@BaseFilename));

   If[Length[FileCopyList] > 0,
      Return[BaseFilename <> ToString[Min[Complement[Table[i, {i, 1, Max[FileCopyList] + 1}], FileCopyList]]]];,
      Return[BaseFilename];
     ];
    ResetDirectory[];
   ];



Options[RunSpectrumGenerator] = {
   InputFilePath -> Global`InputFilePathValue,
   OutputFilePath -> Global`OutputFilePathValue,
   WorkingFolder -> Automatic,
   RemoveTemporaryFiles -> True,
   SpectrumGenerator -> "spheno"(* "softsusy", "suspect" *)};

RunSpectrumGenerator[opt : OptionsPattern[]] := Module[{FlexibleInputFileNames,MoveFileRule, InputFileNameRule, TmpOutputFileNameRule, AddCmdArgs, CurrentWorkingFolder, MoveCommand, RunGeneratorCommand, ChangeDirCommand, RemoveCommentLinesCommand, RemoveTemporaryFilesCommand, ChangeBackToPrevDirCommand},
   
  (* generators require different names of the input file *)
   InputFileNameRule = {
     "spheno" -> "LesHouches.in",
     "softsusy" -> "LesHouches.in",
     "suspect" -> "suspect2_lha.in"};
   
   (* additional command line args in call of spec generators*)
   AddCmdArgs = {
     "spheno" -> "",
     "softsusy" -> "leshouches",
     "suspect" -> ""};
   
   (* temporary output file name of generator *)
   TmpOutputFileNameRule = {
     "spheno" -> "SPheno.spc",
     "softsusy" -> FileNameTake[OptionValue[OutputFilePath]]<>".tmp",
     "suspect" -> "suspect2_lha.out"};
   
   (* some generators cannot treat a path to input file in command line call.
      Thus one has to move input files into current working directory (including renaming...)  *)
   MoveFileRule = {
     "spheno" -> True,
     "softsusy" -> False,
     "suspect" -> True};
   

   If[OptionValue[WorkingFolder] === Automatic,
      CurrentWorkingFolder = DirectoryName[OptionValue[InputFilePath]];,
      CurrentWorkingFolder = OptionValue[WorkingFolder];
     ];

   (* when generator supports flexible input file names,
      generate a unique file name for the temporary output file using the default temporary output file name as basename *)
(*   If[!#[[2]],
      TmpOutputFileNameRule=ReplaceRule[TmpOutputFileNameRule,
                  #[[1]]->GetUniqueFileName[CurrentWorkingFolder,#[[1]]/.TmpOutputFileNameRule]];]&/@MoveFileRule;
*)

   
   (* renaming and moving file to WorkingFolder if needed. 
      otherwise using extended program call to specifiy in and out files *)
   If[OptionValue[SpectrumGenerator] /. MoveFileRule,
      MoveCommand = "cp -f " <> (OptionValue[InputFilePath]) <> " " <> CurrentWorkingFolder <> "/" <> (OptionValue[SpectrumGenerator] /. InputFileNameRule);
      RunGeneratorCommand = ProgramPath[OptionValue[SpectrumGenerator]] <> " " <> (OptionValue[SpectrumGenerator] /. AddCmdArgs) <> "> SLAM.runspec.log 2>&1";,
      MoveCommand = "";
      RunGeneratorCommand = ProgramPath[OptionValue[SpectrumGenerator]] <> " " <> (OptionValue[SpectrumGenerator] /. AddCmdArgs) <> " < " <>OptionValue[InputFilePath] <> " > " <> (OptionValue[SpectrumGenerator] /. TmpOutputFileNameRule);
     ];
   ChangeDirCommand="cd " <> CurrentWorkingFolder<>"; ";
   RemoveCommentLinesCommand="sed '/^#/d' " <> (OptionValue[SpectrumGenerator] /. TmpOutputFileNameRule) <> " > " <> OptionValue[OutputFilePath];
   RemoveTemporaryFilesCommand=If[OptionValue[RemoveTemporaryFiles],
                                  "rm -f "<>(OptionValue[SpectrumGenerator] /. TmpOutputFileNameRule)<>";"
                                   <> If[OptionValue[SpectrumGenerator] /. MoveFileRule,
                                         "rm -f "<> (OptionValue[SpectrumGenerator] /. InputFileNameRule)<>";",
                                         ""],
                                  ""];
   
(* if [ -f SLAM.runspec.log ]; then if [ 0 -lt `stat -c%s SLAM.runspec.log` ]; then echo "Nonempty log file detected";else rm -f SLAM.runspec.log; fi; fi;*)
   


(*   SLAMPrint["MoveCommand: ",MoveCommand];
   SLAMPrint["RunGeneratorCommand: ",RunGeneratorCommand];
   SLAMPrint["RemoveCommentLinesCommand: ",RemoveCommentLinesCommand];
   SLAMPrint["RemoveTemporaryFilesCommand: ",RemoveTemporaryFilesCommand];*)

   
   If[MoveCommand=!="",
      Switch[Run[MoveCommand],
             0,AllFine;,
             _,Message[SLAM::NoFileMove];Return[$Aborted];];
   ];

   Switch[#,
          0,
             AllFine;,
          a_,
             If[FileExistsQ[CurrentWorkingFolder<>"SLAM.runspec.log"],
                Switch[FileByteCount[CurrentWorkingFolder<>"SLAM.runspec.log"],
                       0,Message[SLAM::SpecGenNonZeroExit,#];DeleteFile[CurrentWorkingFolder<>"SLAM.runspec.log"];,
                       a_,Message[SLAM::SpecGenNonZeroExitPlusMessage,#,Import[CurrentWorkingFolder<>"SLAM.runspec.log","Text"]];DeleteFile[CurrentWorkingFolder<>"SLAM.runspec.log"];];,
               Message[SLAM::SpecGenNonZeroExit,#];
             ];
             Return[$Aborted];
         ]&@Run[ChangeDirCommand<>RunGeneratorCommand];

   If[FileExistsQ[CurrentWorkingFolder<>"SLAM.runspec.log"],
   Switch[FileByteCount[CurrentWorkingFolder<>"SLAM.runspec.log"],
          0,DeleteFile[CurrentWorkingFolder<>"SLAM.runspec.log"];,
          a_,Message[SLAM::SpecGenMessage,OptionValue[SpectrumGenerator],Import[CurrentWorkingFolder<>"SLAM.runspec.log","Text"]];DeleteFile[CurrentWorkingFolder<>"SLAM.runspec.log"];Return[$Aborted];];
   ];


   Switch[Run[ChangeDirCommand<>RemoveCommentLinesCommand],
          0,AllFine;,
          _,Message[SLAM::SedError];Return[$Aborted];];

   If[RemoveTemporaryFilesCommand=!="",
      Switch[Run[ChangeDirCommand<>RemoveTemporaryFilesCommand],
             0,AllFine;,
             _,Message[SLAM::NoFileRemove];Return[$Aborted];];
   ];  


   ];




(*For Debug mode ONLY!:(Writes all DebugPrints and warning messages in debug.txt)*)
$DebugReadLesouchesSpectrum = False;
If[$DebugReadLesouchesSpectrum === True, 
  LogFilePathName = "~/debug.txt";
  WriteToLogFile := (LogFile = OpenAppend[LogFilePathName]; WriteString[LogFile, ToString[#] <> "\n"]; Close[LogFile];) &;
  DebugPrint = WriteToLogFile;
  WriteMessageFilePathName = "~/debug.txt";
  WriteMessageToFile := If[Length[Cases[{##}, $Off[], {0, Infinity}]] === 0, 
                           (MessageFile = OpenAppend[WriteMessageFilePathName]; 
                           WriteString[MessageFile, "WARNING: " <> ToString[{##}] <> "\n"];
                           Close[MessageFile];)] &;
                           Unprotect[Message];
                           Message = WriteMessageToFile;];

(* Reading Lesouches Output files supporting arbitrary content in Lesouches Format*)
LSHEntryQ = ((EntryKey /. #) =!= EntryKey) &;
LSHBlockQ = ((BC /. #) =!= BC) &;

LSHReadBEBackup = Flatten[ToExpression[StringCases[#, {
       RegularExpression["^\\s*(\\d+)\\s+(\\d+)\\s+([^# \\s]+)\\s*[#](.*)$"] :> "{EntryKey->{$1,$2},EV->\"$3\",EC->\"$4\"}",
       RegularExpression["^\\s*(\\d+)\\s+([^# \\s]+)\\s*[#](.*)$"] :> "{EntryKey->$1,EV->\"$2\",EC->\"$3\"}",
       RegularExpression["^\\s*(\\d+)\\s+(\\d+)\\s+([^# \\s]+)\\s*$"] :> "{EntryKey->{$1,$2},EV->\"$3\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+([^# \\s]+)\\s*$"] :> "{EntryKey->$1,EV->\"$2\",EC->None}",
       RegularExpression["^\\s*([^# \\s]+)\\s*[#](.*)$"] :> "{EntryKey->None,EV->\"$1\",EC->\"$2\"}",
       RegularExpression["^\\s*([^# \\s]+)\\s*$"] :> "{EntryKey->None,EV->\"$1\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+(.*)$"] :> "{EntryKey->$1,EV->\"$2\",EC->None}"}]]] &;

LSHReadBE = Flatten[ToExpression[StringCases[#, {
       RegularExpression["^\\s*(\\d+)\\s+(\\d+)\\s+([^# \\s]+)\\s*[#]\\s*$"] :> "{EntryKey->{$1,$2},EV->\"$3\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+(\\d+)\\s+([^# \\s]+)\\s*[#]\\s*(.*)$"] :> "{EntryKey->{$1,$2},EV->\"$3\",EC->\"$4\"}",
       RegularExpression["^\\s*(\\d+)\\s+([^# \\s]+)\\s*[#]\\s*$"] :> "{EntryKey->$1,EV->\"$2\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+([^# \\s]+)\\s*[#]\\s*(.*)$"] :> "{EntryKey->$1,EV->\"$2\",EC->\"$3\"}",
       RegularExpression["^\\s*(\\d+)\\s+(\\d+)\\s+([^# \\s]+)\\s*$"] :> "{EntryKey->{$1,$2},EV->\"$3\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+([^# \\s]+)\\s*$"] :> "{EntryKey->$1,EV->\"$2\",EC->None}",
       RegularExpression["^\\s*([^# \\s]+)\\s*[#]\\s*$"] :> "{EntryKey->None,EV->\"$1\",EC->None}",
       RegularExpression["^\\s*([^# \\s]+)\\s*[#]\\s*(.*)$"] :> "{EntryKey->None,EV->\"$1\",EC->\"$2\"}",
       RegularExpression["^\\s*([^# \\s]+)\\s*$"] :> "{EntryKey->None,EV->\"$1\",EC->None}",
       RegularExpression["^\\s*(\\d+)\\s+(.*)$"] :> "{EntryKey->$1,EV->\"$2\",EC->None}"}]]] &;

LSHReadBlockBackup = Flatten[ToExpression[StringCases[#, {
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+(\\w+[^#]*)\\s+[#](.*)$"] :> "{Block->ToUpperCase[\"$1\"],BC->\"$3\",BI->{EV->\"$2\",EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+([^#]+)\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->\"$2\",EC->None}}", 
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+[#](.*)$"] :> "{Block->ToUpperCase[\"$1\"],BC->\"$2\",BI->{EV->None,EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->None,EC->None}}"}]]] &;

LSHReadBlock = Flatten[ToExpression[StringCases[#, {
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+(\\w+[^#]*)\\s+[#]\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->\"$2\",EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+(\\w+[^#]*)\\s+[#]\\s*(.*)$"] :> "{Block->ToUpperCase[\"$1\"],BC->\"$3\",BI->{EV->\"$2\",EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+([^#]+)\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->\"$2\",EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+[#]\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->None,EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s+[#]\\s*(.*)$"] :> "{Block->ToUpperCase[\"$1\"],BC->\"$2\",BI->{EV->None,EC->None}}",
       RegularExpression["[Bb][Ll][Oo][Cc][Kk]\\s+([^# \\s]+)\\s*$"] :> "{Block->ToUpperCase[\"$1\"],BC->None,BI->{EV->None,EC->None}}"}]]] &;

LSHProcessValue[x_, String] := x;
LSHProcessValue[x_, Integer] := ToExpression[x];
LSHProcessValue[x_, Real] := FORTRANToMathematicaNumber[x];

FORTRANToMathematicaNumber = (If[Head[#] =!= String, 
                                 #, 
                                (If[Length[#] === 1, 
                                     #[[1]], 
                                     #] &@ToExpression[StringCases[#, {
         RegularExpression["([+]*[-]*\\d+[.]\\d*)[dDeE]([+]*[-]*\\d+)"] :> "($1)*10^($2)",
         RegularExpression["([+]*[-]*\\d+[.]*\\d*)"] :> "($1)"}]])]/.{}->$MissingData) &;


(* Filters the BlockRule Input for requested  Input and transforms the EVs into the provided format if needed*)
Options[FilterLesHouchesBlockToInputRequest] = {AdoptEntryFormats -> True,
                                                SecureEntryFormats->False,
                                                DataStructure->Global`LesHouchesOutputSQLFormats,
                                                InputRequest -> Global`LesHouchesInputRequest};
FilterLesHouchesBlockToInputRequest[expr_, opt : OptionsPattern[]] := Module[{CurrentBlock, CurrentKey, ProcessFunction},
  
  If[OptionValue[AdoptEntryFormats],
   If[OptionValue[SecureEntryFormats],(* Only needed when original input is given through mathematica input and real numbers have to include FORTRAN rounding bugs*)
      ProcessFunction = LSHProcessValue[ LSHInverseProcessValue[#1,#2],#2]&;
      ,
      ProcessFunction = LSHProcessValue;
    ];,
   ProcessFunction = #1 &
   ];
  
  Which[OptionValue[InputRequest]===All,
     Return[expr];,
     (*Return[Flatten[((CurrentBlock = #[[1]]; (CurrentKey = #[[1]]; (#[[2, 1]] -> ProcessFunction[EV /. ((CurrentKey /. ((CurrentBlock /. expr) /. CurrentBlock -> {}) /. CurrentKey -> {})) /. EV -> $MissingData, #[[2, 2]]])) & /@ #[[2]])) & /@ OptionValue[InputRequest]]];*)
     OptionValue[InputRequest]===AllFormated,
     Return[ReleaseHold[(
                            (
                              (
                               CurrentBlock -> (CurrentBlock /. expr)
                              )/. Rule ->  RulE /. # /. ((RulE[CurrentBlock, CurrentBlock]) -> Seq[]) &@(
                               CurrentBlock = #[[1]]; (
                                                       ((RulE[#[[1]], {RulE[EV, x_], c___}]) :> (Evaluate[RulE[#[[1]], {EV -> Hold[LSHProcessValue[x, #[[2]]]], c}]])) & /@ #[[2]]
                                                      )
                              )
                            ) & /@ OptionValue[DataStructure]
                          )/. RulE -> Rule /. Seq -> Sequence]];,
     True,
     Return[Flatten[((CurrentBlock = #[[1]]; 
                     (CurrentKey = #[[1]];
                      Which[(Head[CurrentKey]===Integer) || (CurrentKey===None),
                     (#[[2, 1]] -> ProcessFunction[EV /. ((CurrentKey /. ((CurrentBlock /. expr) /. CurrentBlock -> {}) /. CurrentKey -> {})) /. {EV -> $MissingData,None->$MissingData}, #[[2, 2]]]),
                      CurrentKey===BI,
                     (#[[2, 1]] -> ProcessFunction[EV /. ((CurrentKey /. ((CurrentBlock /. expr) /. CurrentBlock -> {}) /. CurrentKey -> {})) /. {EV -> $MissingData,None->$MissingData}, #[[2, 2]]]),
                      CurrentKey===BC,
                     (#[[2, 1]] -> ProcessFunction[(CurrentKey /. ((CurrentBlock /. expr) /. CurrentBlock -> {})) /. {CurrentKey -> $MissingData,None->$MissingData}, #[[2, 2]]]),
                      (Length[CurrentKey]===2)&& (Head[CurrentKey[[2]]]===Integer),
                     (#[[2, 1]] -> ProcessFunction[EV/. ((CurrentKey /. ((CurrentBlock /. expr) /. CurrentBlock -> {}) /. CurrentKey -> {})) /. {EV -> $MissingData,None->$MissingData}, #[[2, 2]]]),
                      (Length[CurrentKey]===2)&& (Head[CurrentKey[[2]]]=!= Integer),
                     (#[[2, 1]] -> ProcessFunction[CurrentKey[[2]]/. ((CurrentKey[[1]] /. ((CurrentBlock /. expr) /. CurrentBlock -> {}) /. CurrentKey[[1]] -> {})) /. {CurrentKey[[2]] -> $MissingData,None->$MissingData}, #[[2, 2]]])
                      ]) & /@ #[[2]])) & /@ OptionValue[InputRequest]]];
    ];

];

RenameBlock[BlocknameList_, BlockToBeRenamed_] := Module[{CurrentBlockName, n},
   n = 1;
   CurrentBlockName = BlockToBeRenamed;

   While[MemberQ[BlocknameList, CurrentBlockName],
    CurrentBlockName = BlockToBeRenamed <> ToString[n++];
    ];
   Return[CurrentBlockName];
   ];


(* 
Returns the content of a spectrum file in Mathematica Block Rule notation. 
InputRequest-> All 
  makes it return all entries of the file as string.
InputRequest->LesHouchesInputRequest (or object with similar structure like LesHouchesInputRequest)
  makes it return a filtered list like {MA->2000,Mgl->500,...} where all strings are converted into the given formats (->LesHouchesInputRequest)

*)
Options[ReadLesHouchesSpectrumFile] = {OutputFilePath -> Global`OutputFilePathValue,
                                      Input -> File,
                                      RemoveTemporaryFiles -> True,
                                      DataStructure ->Global`LesHouchesOutputSQLFormats(*Only needed when: InputRequest-> AllFormated *),
                                      InputRequest -> Global`LesHouchesInputRequest(*,All,AllFormated*),
                                      AdoptEntryFormats -> True};
ReadLesHouchesSpectrumFile[opt : OptionsPattern[]] := Module[{LineNumber, ListOfBlocks, LesouchesSpectrumStream, CurrentLine, CurrentBlock, CurrentBE, NewBlock, BlockRule,CurrentKey,BlockNamesList,NewBlockName,RulE},
   
   (* Save names of Blocks in case a block appears more than once in a file like in SPheno outputs...*)
   BlockNamesList={};
   (* Tracks the current number of line*)
   LineNumber=0;

   ListOfBlocks = {};
   If[OptionValue[Input]===File,
      If[FileType[OptionValue[OutputFilePath]]=!=File,
         SLAMPrint["WARNING: Could not read LesHouches file!"];Return[$Aborted];
      ];
      LesouchesSpectrumStream = StringToStream[Import[OptionValue[OutputFilePath], "Text"]];,
      LesouchesSpectrumStream = StringToStream[OptionValue[Input]];
    ];
   (* Getting first line of the file *)
   CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);
   
   (* the first while loop runs until a CurrentBlock has been detected*)
   While[CurrentLine =!= EndOfFile,
    
    (* trying to find block data *)
    CurrentBlock = LSHReadBlock[CurrentLine];
    
    If[LSHBlockQ[CurrentBlock],
       AppendTo[BlockNamesList,Block/.CurrentBlock];
       DebugPrint["L" <> ToString[LineNumber] <> ": First Block detected!  " <> ToString[CurrentBlock]];
     (* when the first block has been detected the while loop fills this and following blocks with entries *)
     
       CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);
       While[CurrentLine =!= EndOfFile,
             Which[CurrentBE = LSHReadBE[CurrentLine]; DebugPrint["  EntryCeck: " <> ToString[CurrentBE]]; LSHEntryQ[CurrentBE],
                    DebugPrint["L" <> ToString[LineNumber] <> ":Current Line -> Entry  " <> ToString[CurrentBE]];
                    AppendTo[CurrentBlock, CurrentBE];
                    CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);,
                   NewBlock = LSHReadBlock[CurrentLine]; DebugPrint["  BlockCeck: " <> ToString[NewBlock]]; LSHBlockQ[NewBlock],
                    DebugPrint["L" <> ToString[LineNumber] <> ":Current Line ->  Block  " <> ToString[CurrentBlock]];
                    AppendTo[ListOfBlocks, CurrentBlock];
                    (* when a block name appears for the second time rename the new block! *)
                    If[MemberQ[BlockNamesList, Block/.NewBlock],
                       NewBlockName=RenameBlock[BlockNamesList, Block/.NewBlock];
                       AppendTo[BlockNamesList, NewBlockName];
                       NewBlock=NewBlock/.((Block/.NewBlock)->NewBlockName);,
                       AppendTo[BlockNamesList, Block/.NewBlock];
                    ];
                    CurrentBlock = NewBlock;
                    CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);,
                   True,
                    DebugPrint["L" <> ToString[LineNumber] <> ":Current Line -> NOMATCH 2" <> ToString[CurrentLine]];
                    CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);
                   ];
      ];

      AppendTo[ListOfBlocks, CurrentBlock];,
       DebugPrint["L" <> ToString[LineNumber] <> ":Current Line -> NOMATCH 1" <> ToString[CurrentLine]];
     ];
    (* Get a new line when no CurrentBlock has been found *)
    
    CurrentLine = (LineNumber++; Read[LesouchesSpectrumStream, String]);
   ];
   
   Close[LesouchesSpectrumStream];
   
   BlockRule = (Block /. Take[#, 1]) -> Join[Take[#, {2, 3}], ((EntryKey /. #) -> Drop[#, 1]) & /@ Drop[#, 3]] & /@ ListOfBlocks;

   (* Treating Spectrum generator dependend output: spheno uses to write two gauge blocks... only the second one is of interest*)
         If[Position[BlockRule, "GAUGE1",{2}]=!={},
            BlockRule=BlockRule/.{Rule["GAUGE", x_] :> Seq[]} /. Seq -> Sequence /. Rule["GAUGE1", x_] :> Rule["GAUGE", x];
         ];



   If[OptionValue[RemoveTemporaryFiles],
      DebugPrint["Removing temporary file: ",OptionValue[OutputFilePath]];
      DeleteFile[OptionValue[OutputFilePath]];
   ];


   (* Checking for error and warning messages in SLHA output file *)
   If[!FreeQ[BlockRule,"SPINFO"],
   ((If[Head[#[[1]]] === String, Message[SLAM::SpecGenWarning,#[[1]]]]; 
     If[Head[#[[2]]] === String, Message[SLAM::SpecGenError,#[[2]]]];) &@(EV /. # & /@ ({3, 4} /. ("SPINFO" /. BlockRule)/. {3 -> {(*No Warning*)}, 4 -> {(*No Error*)}})));
   ];
   
   Which[OptionValue[InputRequest] === All,
      Return[BlockRule];,
      (* Filtering *)True,
      Return[FilterLesHouchesBlockToInputRequest[BlockRule, AdoptEntryFormats -> OptionValue[AdoptEntryFormats], InputRequest->OptionValue[InputRequest], DataStructure ->OptionValue[DataStructure]]];
    ];
   
   
   ];


(* # BEGIN OF SQL Data Base Interface # *)

(* transforms an EntryKey like 1 or {1,2} into "1" or "1_2" *)
ToSQLEntryString = Which[Head[#] === List, 
                         StringReplace[ToString[#], {"," -> "_", "}" -> "", "{" -> "", RegularExpression["\\s+"] :> ""}],
                         #===None,
                         "NONE",
                         True,
                         ToString[#]] &;

(* SQL data type definition *)
SQLDataTypeName[Real] = "DOUBLE";
SQLDataTypeName[Integer] = "INTEGER";
SQLDataTypeName[String] = "VARCHAR";

(* Transforms Matheamtica BlockRule into SQL Data Base Lists *)
(* Examples:

"BLOCKNAME"->{1->{EV->20,EC->"test2"}, BI->"info",BC-> {EV->"test",EC->None}}

will generate (Output -> All)

{
{"BLOCKNAME_1","INTEGER",20},
{"BLOCKNAME_1_COMMENT","VARCHAR","test2"},
{"BLOCKNAME_INFO","VARCHAR","info"},
{"BLOCKNAME_COMMENT","VARCHAR","test"},
}

"BLOCKNAME"->{{1,2}->{EV->20,EC->"test2"}}

results in (Output -> All)

{
{"BLOCKNAME_1_2","INTEGER",20},
{"BLOCKNAME_1_2_COMMENT","VARCHAR","test2"},
}

*)
Options[ToSQLLesHouchesData] = {DataStructure -> Global`LesHouchesInputSQLFormats,
                                NoComments->False,
                                AdoptEntryFormats->True,
                                SecureEntryFormats->False,
                                Direction->In,(*,Out  In:-> Prefix="I_" (I)nput for generators Out:-> Prefix="O_" (O)utput of generators*)
                                Output -> All(*,Declarations,Definition*)};
ToSQLLesHouchesData[LeshouchesData_, opt:OptionsPattern[]] := Module[{ProcessFunction,Prefix,CurrentBlock, CurrentEntry, CurrentDataType, BlockDetails, BlockEntries},
   

  If[OptionValue[AdoptEntryFormats],
   If[OptionValue[SecureEntryFormats],(* Only needed when original input is given through mathematica input and real numbers have to include FORTRAN rounding bugs*)
      ProcessFunction = LSHProcessValue[ LSHInverseProcessValue[#1,#2],#2]&;
      ,
      ProcessFunction = LSHProcessValue;
    ];,
   ProcessFunction = #1 &
   ];



 Which[OptionValue[Direction]===In,
       Prefix="I_";,
       OptionValue[Direction]===Out,
       Prefix="O_";,
       Head[OptionValue[Direction]]===String,
       Prefix=OptionValue[Direction];,
       True,
       Prefix="";
  ];
 If[OptionValue[NoComments],
    BlockDetails = Flatten[(CurrentBlock = #[[1]]; 
                            {{Prefix<>CurrentBlock <> "_INFO", SQLDataTypeName[String], ToString[(EV /. #2) /. EV -> $MissingData]}} &@(Sequence @@(({BC, BI} /. ((CurrentBlock /. LeshouchesData) /. CurrentBlock -> {})) /. {BC -> $MissingData, BI -> (EV -> $MissingData)}))) & /@ OptionValue[DataStructure], 1];
    BlockEntries = Flatten[(CurrentBlock = #[[1]]; 
                          (CurrentEntry = #[[1]]; 
                           CurrentDataType = #[[2]]; 
              {{Prefix<>CurrentBlock <> "_" <> ToSQLEntryString[CurrentEntry], 
                SQLDataTypeName[CurrentDataType], 
                ProcessFunction[#1, CurrentDataType]}} &@(Sequence @@ (({EV, EC} /. (Replace[CurrentEntry /. Replace[(CurrentBlock/. LeshouchesData), CurrentBlock -> {},{0}], CurrentEntry -> {},{0}])) /. {EV -> $MissingData, EC -> $MissingData}))) & /@ #[[2]]) & /@ OptionValue[DataStructure], 2];,
    BlockDetails = Flatten[(CurrentBlock = #[[1]]; 
                            {{Prefix<>CurrentBlock <> "_COMMENT", SQLDataTypeName[String], ToString[#1]}, {Prefix<>CurrentBlock <> "_INFO", SQLDataTypeName[String], ToString[(EV /. #2) /. EV -> $MissingData]}} &@(Sequence @@(({BC, BI} /. ((CurrentBlock /. LeshouchesData) /. CurrentBlock -> {})) /. {BC -> $MissingData, BI -> (EV -> $MissingData)}))) & /@ OptionValue[DataStructure], 1];
    BlockEntries = Flatten[(CurrentBlock = #[[1]]; 
                          (CurrentEntry = #[[1]]; 
                           CurrentDataType = #[[2]]; 
              {{Prefix<>CurrentBlock <> "_" <> ToSQLEntryString[CurrentEntry], 
                SQLDataTypeName[CurrentDataType], 
                ProcessFunction[#1, CurrentDataType]}, 
              {Prefix<>CurrentBlock <>"_" <> ToSQLEntryString[CurrentEntry] <> "_COMMENT", 
               SQLDataTypeName[String], #2}} &@(Sequence @@ (({EV, EC} /. (Replace[CurrentEntry /. Replace[(CurrentBlock/. LeshouchesData),CurrentBlock -> {},{0}], CurrentEntry -> {},{0}]) /. {EV -> $MissingData, EC -> $MissingData})))) & /@ #[[2]]) & /@ OptionValue[DataStructure], 2];
   ];
   Which[OptionValue[Output] === All,
         Return[Join[BlockEntries, BlockDetails]];,
         OptionValue[Output] === Declarations,
         Return[{#[[1]], #[[2]]} & /@ Join[BlockEntries, BlockDetails]];,
         OptionValue[Output] === Definitions,
         Return[(Which[#[[3]] === None, 
                       Seq[], 
                       #[[3]] === "None", 
                       Seq[], 
                       #[[3]] === $MissingData, 
                       Seq[], 
                       #[[3]] === "$MissingData", 
                       Seq[], 
                       True, 
                       {#[[1]], #[[3]]}] & /@ Join[BlockEntries, BlockDetails]) /. Seq -> Sequence];
    ];
];

Options[CreateSQLLesHouchesDataBase] = {SQLDataBaseType->Global`SQLDataBaseTypeValue,
                                        SQLDataBaseOptions->Global`SQLDataBaseOptionsValue,  
                                        SQLDataBase ->Global`SQLDataBaseValue,
                                        TableName -> "SPECTRATABLE",
                                        TableDeclarationList -> {(*is defined below...*)},
                                        KeepSQLConnection -> True,
                                        MaxSQLConnectionAttempts->10,
                                        SQLVarCharLength->Global`SQLVarCharLengthValue
                                       };


CreateSQLLesHouchesDataBase[opt : OptionsPattern[]] := Module[{CurrentSQLConnection,ConnectionAttempts},
   
   If[Head[$LesHouchesSQLConnectionList] === Symbol, 
      $LesHouchesSQLConnectionList = {};];
   
   If[Length[Names["DatabaseLink`*"]] < 20, 
      SLAMPrint["Loading DatabaseLink` Package..."];Needs["DatabaseLink`"];];
   
   
   Which[(Length[$LesHouchesSQLConnectionList]=!=0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection=$LesHouchesSQLConnectionList[[1]];,
         (Length[$LesHouchesSQLConnectionList]===0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ConnectionAttempts = 0;
         While[(CurrentSQLConnection === $Failed) && (ConnectionAttempts < OptionValue[MaxSQLConnectionAttempts]),
               Pause[RandomReal[{1, 10}]];
               ConnectionAttempts++;
               CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ];
         If[CurrentSQLConnection===$Failed,Message[SLAM::NoDBConnect];Abort[];];
         AppendTo[$LesHouchesSQLConnectionList, CurrentSQLConnection];,
         Head[OptionValue[SQLDataBase]] === DatabaseLink`SQLConnection,
         CurrentSQLConnection = OptionValue[SQLDataBase];,
         (Head[OptionValue[SQLDataBase]] === Integer) && (OptionValue[SQLDataBase] > 0) && (Length[$LesHouchesSQLConnectionList] >= OptionValue[SQLDataBase]),
         CurrentSQLConnection = $LesHouchesSQLConnectionList[[OptionValue[SQLDataBase]]];,
         True,
         Return[$Aborted];
        ];
   
   DatabaseLink`SQLCreateTable[CurrentSQLConnection, OptionValue[TableName], Append[(If[#[[2]]==="VARCHAR",
                                                                                        DatabaseLink`SQLColumn[#[[1]], "DataTypeName" -> #[[2]], "DataLength" -> OptionValue[SQLVarCharLength]],
                                                                                        DatabaseLink`SQLColumn[#[[1]], "DataTypeName" -> #[[2]], "Nullable" -> True]] & /@ OptionValue[TableDeclarationList]), 
                                                                                     DatabaseLink`SQLColumn["TIMESTAMP", "DataTypeName" -> "TIMESTAMP"]]];
  (* 
  Global`TABLEDECCOMMAND=My`SQLCreateTable[CurrentSQLConnection, OptionValue[TableName], Append[(If[#[[2]]==="VARCHAR",
                                                                                        DatabaseLink`SQLColumn[#[[1]], "DataTypeName" -> #[[2]], "DataLength" -> OptionValue[SQLVarCharLength]],
                                                                                        DatabaseLink`SQLColumn[#[[1]], "DataTypeName" -> #[[2]], "Nullable" -> True]] & /@ OptionValue[TableDeclarationList]),
                                                                                     DatabaseLink`SQLColumn["TIMESTAMP", "DataTypeName" -> "TIMESTAMP"]]];
  *)
 
   If[!OptionValue[KeepSQLConnection], 
       DatabaseLink`CloseSQLConnection[CurrentSQLConnection];
       $LesHouchesSQLConnectionList = ($LesHouchesSQLConnectionList /.(CurrentSQLConnection -> Seq[])) /. Seq -> Sequence;];
   ];



Options[WriteSQLLesHouchesData] = {SQLDataBaseType->Global`SQLDataBaseTypeValue,
                                   SQLDataBaseOptions->Global`SQLDataBaseOptionsValue, 
                                   SQLDataBase -> Global`SQLDataBaseValue,
                                   TableName -> "SPECTRATABLE",
                                   TableDefinitionList -> SQLLesHouchesTestDefinitions,
                                   KeepSQLConnection -> True,
                                   MaxSQLConnectionAttempts->10};

WriteSQLLesHouchesData[opt : OptionsPattern[]] := Module[{CurrentSQLConnection,ConnectionAttempts},
  
  If[Head[$LesHouchesSQLConnectionList] === 
     Symbol, $LesHouchesSQLConnectionList = {};];
  
  If[Length[Names["DatabaseLink`*"]] < 20, 
     SLAMPrint["Loading DatabaseLink` Package..."];Needs["DatabaseLink`"];];
  
   Which[(Length[$LesHouchesSQLConnectionList]=!=0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection=$LesHouchesSQLConnectionList[[1]];,
         (Length[$LesHouchesSQLConnectionList]===0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ConnectionAttempts = 0;
         While[(CurrentSQLConnection === $Failed) && (ConnectionAttempts < OptionValue[MaxSQLConnectionAttempts]),
               Pause[RandomReal[{1, 10}]];
               ConnectionAttempts++;
               CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]],If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ];
         If[CurrentSQLConnection===$Failed,Message[SLAM::NoDBConnect];Abort[];];
         AppendTo[$LesHouchesSQLConnectionList, CurrentSQLConnection];,
         Head[OptionValue[SQLDataBase]] === DatabaseLink`SQLConnection,
         CurrentSQLConnection = OptionValue[SQLDataBase];,
         (Head[OptionValue[SQLDataBase]] === Integer) && (OptionValue[SQLDataBase] > 0) && (Length[$LesHouchesSQLConnectionList] >= OptionValue[SQLDataBase]),
         CurrentSQLConnection = $LesHouchesSQLConnectionList[[OptionValue[SQLDataBase]]];,
         True,
         Return[$Aborted];
        ];
  
  
  DatabaseLink`SQLInsert[CurrentSQLConnection, OptionValue[TableName], Sequence @@ Transpose[Append[OptionValue[TableDefinitionList], {"TIMESTAMP", DatabaseLink`SQLDateTime[Date[]]}]]];
  
  If[!OptionValue[KeepSQLConnection], DatabaseLink`CloseSQLConnection[CurrentSQLConnection];
     $LesHouchesSQLConnectionList = ($LesHouchesSQLConnectionList /.(CurrentSQLConnection -> Seq[])) /. Seq -> Sequence;];
  ];


Options[DeleteSQLLesHouchesTable] = {SQLDataBaseType->Global`SQLDataBaseTypeValue,
                                     SQLDataBaseOptions->Global`SQLDataBaseOptionsValue,                                  
                                     SQLDataBase -> Global`SQLDataBaseValue,
                                     TableName -> "SPECTRATABLE",
                                     KeepSQLConnection -> True,
                                     MaxSQLConnectionAttempts->10,
                                     Condition -> Sequence[]};

DeleteSQLLesHouchesTable[opt : OptionsPattern[]] := Module[{CurrentSQLConnection,ConnectionAttempts},
   
   If[Head[$LesHouchesSQLConnectionList] === Symbol, 
      $LesHouchesSQLConnectionList = {};];
   
   If[Length[Names["DatabaseLink`*"]] < 20, 
      SLAMPrint["Loading DatabaseLink` Package..."];Needs["DatabaseLink`"];];
   
   Which[(Length[$LesHouchesSQLConnectionList]=!=0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection=$LesHouchesSQLConnectionList[[1]];,
         (Length[$LesHouchesSQLConnectionList]===0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ConnectionAttempts = 0;
         While[(CurrentSQLConnection === $Failed) && (ConnectionAttempts < OptionValue[MaxSQLConnectionAttempts]),
               Pause[RandomReal[{1, 10}]];
               ConnectionAttempts++;
               CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ];
         If[CurrentSQLConnection===$Failed,Message[SLAM::NoDBConnect];Abort[];];
         AppendTo[$LesHouchesSQLConnectionList, CurrentSQLConnection];,
         Head[OptionValue[SQLDataBase]] === DatabaseLink`SQLConnection,
         CurrentSQLConnection = OptionValue[SQLDataBase];,
         (Head[OptionValue[SQLDataBase]] === Integer) && (OptionValue[SQLDataBase] > 0) && (Length[$LesHouchesSQLConnectionList] >= OptionValue[SQLDataBase]),
         CurrentSQLConnection = $LesHouchesSQLConnectionList[[OptionValue[SQLDataBase]]];,
         True,
         Return[$Aborted];
        ];
   
   DatabaseLink`SQLDelete[CurrentSQLConnection, OptionValue[TableName], OptionValue[Condition]];
   
   If[!OptionValue[KeepSQLConnection], 
      DatabaseLink`CloseSQLConnection[CurrentSQLConnection];
      $LesHouchesSQLConnectionList = ($LesHouchesSQLConnectionList /.(CurrentSQLConnection -> Seq[])) /. Seq -> Sequence;];
   
   ];


(* Removing a SQL table competely *)
Options[DropSQLLesHouchesTable] = {SQLDataBaseType->Global`SQLDataBaseTypeValue,
                                     SQLDataBaseOptions->Global`SQLDataBaseOptionsValue,                                  
                                     SQLDataBase -> Global`SQLDataBaseValue,
                                     TableName -> "TEST",
                                     KeepSQLConnection -> True,
                                     MaxSQLConnectionAttempts->10};

DropSQLLesHouchesTable[opt : OptionsPattern[]] := Module[{CurrentSQLConnection,ConnectionAttempts},
   
   If[Head[$LesHouchesSQLConnectionList] === Symbol, 
      $LesHouchesSQLConnectionList = {};];
   
   If[Length[Names["DatabaseLink`*"]] < 20, 
      SLAMPrint["Loading DatabaseLink` Package..."];Needs["DatabaseLink`"];];
   
   Which[(Length[$LesHouchesSQLConnectionList]=!=0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection=$LesHouchesSQLConnectionList[[1]];,
         (Length[$LesHouchesSQLConnectionList]===0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ConnectionAttempts = 0;
         While[(CurrentSQLConnection === $Failed) && (ConnectionAttempts < OptionValue[MaxSQLConnectionAttempts]),
               Pause[RandomReal[{1, 10}]];
               ConnectionAttempts++;
               CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ];
         If[CurrentSQLConnection===$Failed,Message[SLAM::NoDBConnect];Abort[];];
         AppendTo[$LesHouchesSQLConnectionList, CurrentSQLConnection];,
         Head[OptionValue[SQLDataBase]] === DatabaseLink`SQLConnection,
         CurrentSQLConnection = OptionValue[SQLDataBase];,
         (Head[OptionValue[SQLDataBase]] === Integer) && (OptionValue[SQLDataBase] > 0) && (Length[$LesHouchesSQLConnectionList] >= OptionValue[SQLDataBase]),
         CurrentSQLConnection = $LesHouchesSQLConnectionList[[OptionValue[SQLDataBase]]];,
         True,
         Return[$Aborted];
        ];
   
   DatabaseLink`SQLDropTable[CurrentSQLConnection, OptionValue[TableName]];
   
   If[!OptionValue[KeepSQLConnection], 
      DatabaseLink`CloseSQLConnection[CurrentSQLConnection];
      $LesHouchesSQLConnectionList = ($LesHouchesSQLConnectionList /.(CurrentSQLConnection -> Seq[])) /. Seq -> Sequence;];
   
   ];



(* Converts user input request to a list of column names written in the output
  For e.g.:
  {"MASS" -> {BI -> {X1, Real}, 
              1 -> {X2, Integer}, 
              BC -> {X3, String}, 
             {2, EV} -> {X4, Real}, 
             {2, EC} -> {X5,String}}, 
    "SOFT" -> {BC -> {X6, String}, 
             None -> {X7, String}}}
becomes:
 {"O_MASS_INFO", "O_MASS_1", "O_MASS_COMMENT", "O_MASS_2", "O_MASS_2_COMMENT", "O_SOFT_COMMENT", "O_SOFT_NONE"}

 *)
Options[InputRequestToSQLSelectList] = {Prefix -> "O_"};
InputRequestToSQLSelectList[inputrequest_, opt : OptionsPattern[]] := Module[{Blockname, KeyEntry},
   Return[Flatten[#, 1] &@Reap[(Blockname = ToUpperCase[#[[1]]];
           (KeyEntry = #[[1]];
              Which[Head[KeyEntry] === Integer,
                    Sow[OptionValue[Prefix] <> Blockname <> "_" <> ToString[#[[1]]], 1];,
                   (Head[KeyEntry] === List) && (Head[KeyEntry[[1]]] === Integer) && (KeyEntry[[2]] === EV),
                    Sow[OptionValue[Prefix] <> Blockname <> "_" <> ToString[#[[1, 1]]], 1];,
                   (Head[KeyEntry] === List) && (Head[KeyEntry[[1]]] === Integer) && (Head[KeyEntry[[2]]] === Integer),
                    Sow[OptionValue[Prefix] <> Blockname <> "_" <> ToString[#[[1, 1]]]<> "_" <> ToString[#[[1, 2]]], 1];,
                    KeyEntry === None,
                    Sow[OptionValue[Prefix] <> Blockname <> "_NONE", 1];,
                    KeyEntry === BI,
                    Sow[OptionValue[Prefix] <> Blockname <> "_INFO", 1];,
                    KeyEntry === BC,
                    Sow[OptionValue[Prefix] <> Blockname <> "_COMMENT", 1];,
                    (Head[KeyEntry] === List) && (Head[KeyEntry[[1]]] === Integer) && (KeyEntry[[2]] === EC),
                    Sow[OptionValue[Prefix] <> Blockname <> "_" <> ToString[#[[1, 1]]] <> "_COMMENT", 1];
                   ];) & /@ #[[2]]) & /@ (inputrequest), 1][[2]]];
   ];




(* Converting SQL DataBase Output to BlockRule notation in Mathematica *)
Options[ReadSQLLesHouchesEntry] = {BlockPrefix -> "I_"};
ReadSQLLesHouchesEntry[opt : OptionsPattern[]] = ToExpression[If[Length[#] > 0, #[[1]], #] &@StringCases[#, {
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_]NONE$"] :> "\"$1\"->{None->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_]NONE[_]COMMENT$"] :> "\"$1\"->{None->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_](\\d+)[_](\\d+)$"] :> "\"$1\"->{{$2,$3}->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_](\\d+)[_](\\d+)[_]COMMENT$"] :> "\"$1\"->{{$2,$3}->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_](\\d+)$"] :> "\"$1\"->{$2->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_](\\d+)[_]COMMENT$"] :> "\"$1\"->{$2->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_]COMMENT$"] :> "\"$1\"->{BC->#}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(\\w+)[_]INFO$"] :> "\"$1\"->{BI\[Rule]{EV->#,EC->None}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(.*)$"] :> "\"$1\"->{EV->#}&"}]] &;

ReadSQLLesHouchesEntryBackup[opt : OptionsPattern[]] = ToExpression[If[Length[#] > 0, #[[1]], #] &@StringCases[#, {
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_]NONE$"] :> "\"$1\"->{None->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_]NONE[_]COMMENT$"] :> "\"$1\"->{None->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_](\\d+)[_](\\d+)$"] :> "\"$1\"->{{$2,$3}->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_](\\d+)[_](\\d+)[_]COMMENT$"] :> "\"$1\"->{{$2,$3}->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_](\\d+)$"] :> "\"$1\"->{$2->{EV->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_](\\d+)[_]COMMENT$"] :> "\"$1\"->{$2->{EC->#}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_]COMMENT$"] :> "\"$1\"->{BC->#}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "([^_]+)[_]INFO$"] :> "\"$1\"->{BI\[Rule]{EV->#,EC->None}}&",
       RegularExpression["^" <> OptionValue[BlockPrefix] <> "(.*)$"] :> "\"$1\"->{EV->#}&"}]] &;



(* Merges a List of RuleLists 
Example:

{{"BLOCK"->{1->{EV->12}}},{"BLOCK"->{1->{EC->"comment"}}}}
->
{"BLOCK"->{1->{EV->12,EC->"comment"}}}
*)
MergeRulesList[expr_] := Which[(Length[expr] === 1) && (Head[expr] === List) && (Head[expr[[1]]] ===Rule),
                               DebugPrint["1", expr]; Return[{expr[[1, 1]] -> MergeRulesList[expr[[1, 2]]]}];,
                               (Length[expr] === 2) && (Head[expr] === Rule),
                               DebugPrint["2", expr]; Return[expr[[1]] -> MergeRulesList[expr[[2]]]];,
                               (Length[expr] > 1) && (Head[expr] === List) && (Head[expr[[1]]] === Rule),
                               DebugPrint["3", expr]; Return[(#[[1, 1]] -> MergeRulesList[If[Length[#] > 1, Join @@ #, # /. List[List[a___]] :> List[a]] &@(#[[2]] & /@ #)]) & /@ Gather[expr, (#1[[1]] === #2[[1]]) &]];,
                               (Length[expr] === 1) && (Head[expr] === List) && (Length[expr[[1]]] === 0),
                               DebugPrint["4", expr]; Return[expr[[1]]],
                               True,
                               DebugPrint["5", expr]; Return[expr];
                              ];

Options[SQLLesHouchesDataToMathematica]={BlockPrefix->"I_"};
SQLLesHouchesDataToMathematica[expr_,opt:OptionsPattern[]] := ((*SLAMPrint["2e T=",SessionTime[]-StartingSessionTime];*)((*SLAMPrint["2f T=",SessionTime[]-StartingSessionTime];*)MergeRulesList[#]) &@(ReadSQLLesHouchesEntry[BlockPrefix->OptionValue[BlockPrefix]][#[[1]]][#[[2]] /. {Null -> $MissingData, "Null" -> $MissingData}] & /@ expr));

Options[SelectSQLLesHouchesTable] = {SQLDataBaseType->Global`SQLDataBaseTypeValue,
                                     SQLDataBaseOptions->Global`SQLDataBaseOptionsValue, 
                                     SQLDataBase -> Global`SQLDataBaseValue,
                                     TableName -> "SPECTRATABLE",
                                     KeepSQLConnection -> True,
                                     MaxSQLConnectionAttempts->10,
                                     InputRequest->Global`LesHouchesInputRequest,
                                     LoadCompleteSpectra->False,
                                     CheckList -> SQLLesHouchesTestDefinitions,
                                     OutputRequestList -> {(*Is defined below *)},
                                     Direction->Out,
                                     OutputFormat -> LesHouchesBlockRule(*TableDefinitionList,LesHouchesBlockRule,PureValueList*)};

SelectSQLLesHouchesTable[opt : OptionsPattern[]] := Module[{Prefix ,CurrentSQLConnection, SelectResult, SelectList, n,ConnectionAttempts,UserOutputRequestList},

   (* 
     If one wants to obtain Input Data from Data Base when using OutputFormat -> LesHouchesBlockRule 
     set OptionValue[Direction]-> In
   *)
    Which[OptionValue[Direction]===In,
       Prefix="I_";,
       OptionValue[Direction]===Out,
       Prefix="O_";,
       Head[OptionValue[Direction]]===String,
       Prefix=OptionValue[Direction];,
       True,
       Prefix="";
    ]; 
(*    SLAMPrint["2a T=",SessionTime[]-StartingSessionTime]; *)
   If[Head[$LesHouchesSQLConnectionList] === Symbol, 
      $LesHouchesSQLConnectionList = {};];
   
(*   If[Length[Names["DatabaseLink`*"]] < 20, 
      Print["Loading DatabaseLink` Package..."];Needs["DatabaseLink`"];];*)

   Needs["DatabaseLink`"];
   
   Which[(Length[$LesHouchesSQLConnectionList]=!=0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection=$LesHouchesSQLConnectionList[[1]];,
         (Length[$LesHouchesSQLConnectionList]===0)&&(Head[OptionValue[SQLDataBase]] === String),
         CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ConnectionAttempts = 0;
         While[(CurrentSQLConnection === $Failed) && (ConnectionAttempts < OptionValue[MaxSQLConnectionAttempts]),
               Pause[RandomReal[{1, 10}]];
               ConnectionAttempts++;
               CurrentSQLConnection = DatabaseLink`OpenSQLConnection[DatabaseLink`JDBC[OptionValue[SQLDataBaseType], OptionValue[SQLDataBase]], If[Head[#]===List,Sequence@@#,#]&@OptionValue[SQLDataBaseOptions]];
         ];
         If[CurrentSQLConnection===$Failed,Message[SLAM::NoDBConnect];Abort[];];
         AppendTo[$LesHouchesSQLConnectionList, CurrentSQLConnection];,
         Head[OptionValue[SQLDataBase]] === DatabaseLink`SQLConnection,
         CurrentSQLConnection = OptionValue[SQLDataBase];,
         (Head[OptionValue[SQLDataBase]] === Integer) && (OptionValue[SQLDataBase] > 0) && (Length[$LesHouchesSQLConnectionList] >= OptionValue[SQLDataBase]),
         CurrentSQLConnection = $LesHouchesSQLConnectionList[[OptionValue[SQLDataBase]]];,
         True,
         Return[$Aborted];
        ];
(*    SLAMPrint["2b T=",SessionTime[]-StartingSessionTime]; *)
   (* Only load full spectra output, when either this is forced by LoadCompleteSpectra->True or user requested all data with All or AllFormated*)
   Which[OptionValue[LoadCompleteSpectra] || (OptionValue[InputRequest]===All)||(OptionValue[InputRequest]===AllFormated),
         SelectList = Which[Depth[#] === 3, 
                      (#[[1]]) & /@ #, 
                      Depth[#] === 2, 
                      #, 
                      Depth[#] === 1, 
                      {#}] &@OptionValue[OutputRequestList];,
        True,
        UserOutputRequestList=InputRequestToSQLSelectList[OptionValue[InputRequest]];
        If[Length[Complement[UserOutputRequestList,OptionValue[OutputRequestList]]]=!=0,Message[SLAM::IllegalDataRequest];Abort[];];
        SelectList =UserOutputRequestList;
   ];
(*    Global`MySelectList=SelectList; *)
(*    SLAMPrint["2c T=",SessionTime[]-StartingSessionTime]; *)
(*   SelectResultCOMMAND=(My`SQLSelect[CurrentSQLConnection, OptionValue[TableName], SelectList,(*(SLAMPrint["2cI T=",SessionTime[]-StartingSessionTime]);*) And @@ ((*(SLAMPrint["2cII T=",SessionTime[]-StartingSessionTime]);*)((DatabaseLink`SQLColumn[#[[1]]] == #[[2]]) & /@ OptionValue[CheckList]))]);*)
   SelectResult = (DatabaseLink`SQLSelect[CurrentSQLConnection, OptionValue[TableName], SelectList,(*(SLAMPrint["2cI T=",SessionTime[]-StartingSessionTime]);*) And @@ ((*(SLAMPrint["2cII T=",SessionTime[]-StartingSessionTime]);*)((DatabaseLink`SQLColumn[#[[1]]] == #[[2]]) & /@ OptionValue[CheckList]))]);
 (* Global`MySelectResult=SelectResult;*)
(*    SLAMPrint["2d T=",SessionTime[]-StartingSessionTime]; *)
   If[!OptionValue[KeepSQLConnection], 
      DatabaseLink`CloseSQLConnection[CurrentSQLConnection];
      $LesHouchesSQLConnectionList = ($LesHouchesSQLConnectionList /.(CurrentSQLConnection -> Seq[])) /. Seq -> Sequence;];
   
   Which[OptionValue[OutputFormat] === PureValueList,
         Return[(SelectResult) /. List[List[c___]] :> List[c]];,
         OptionValue[OutputFormat] === TableDefinitionList,
         Return[(Transpose[{SelectList, #}] & /@ SelectResult) /.List[List[c___]] :> List[c]];,
         OptionValue[OutputFormat] === LesHouchesBlockRule,
         Return[(SQLLesHouchesDataToMathematica[Transpose[{SelectList, #}],BlockPrefix->Prefix] & /@ SelectResult)];
        ];
   ];


(* Setting up SQL declaration list *)
SetOptions[CreateSQLLesHouchesDataBase, TableDeclarationList -> (Join[ToSQLLesHouchesData[$Aborted -> $Failed,
                                                    DataStructure -> Global`LesHouchesInputSQLFormats,
                                                    NoComments -> False,
                                                    Direction -> In,
                                                    Output -> Declarations],
                                ToSQLLesHouchesData[$Aborted -> $Failed,
                                                    DataStructure -> Global`LesHouchesOutputSQLFormats,
                                                    NoComments -> False,
                                                    Direction -> Out,
                                                    Output -> Declarations]])];
(* Setting up SQL output request with or without comments *)
SetOptions[SelectSQLLesHouchesTable, OutputRequestList -> (Sort[#[[1]] & /@ ToSQLLesHouchesData[$Aborted -> $Failed,
                                                                      DataStructure -> Global`LesHouchesOutputSQLFormats,
                                                                      NoComments -> False,(* Speedup when True, but no access to comments. New selection method -> keep False!*)
                                                                      Direction -> Out,
                                                                      Output -> Declarations]])];


ModelGotNoModselQ = If[(Length["MODSEL" /. #] > 0 || Length["modsel" /. #] > 0 || Length["Modsel" /. #] > 0), False, True] &;


(* Removing irrelevant hard set options from list *)
Options[ObtainLesHouchesSpectrum] = Complement[#,FilterRules[#,{DataStructure,Direction,CheckList,OutputRequestList,SecureEntryFormats,OutputFormat,AdoptEntryFormats,AppendToFile}]]&@Union[{UseDataBase->True,
                                           Silent->False,
                                           ExtendDataBase -> True,
                                           RefreshDataBase -> False,
                                           ClearDataBase -> False(*True,Only*)},
                                           Options[CreateSQLLesHouchesDataBase], 
                                           Options[WriteSpectrumGeneratorInput],
                                           Options[SelectSQLLesHouchesTable],
                                           Options[FilterLesHouchesBlockToInputRequest],
                                           Options[RunSpectrumGenerator]];
ObtainLesHouchesSpectrum[opt : OptionsPattern[]] := Module[{Status,GeneratorInputString, GeneratorInputBlockRule, GeneratorInputSQLDefinitions, GeneratorOutputSQLDefinitions, SQLDataBaseOutput, NewSpectrumBlockRule},
  
(*   StartingSessionTime=SessionTime[]; *)
  If[!OptionValue[Silent],Status=SLAMPrint;];
  If[(OptionValue[ClearDataBase] === True) || (OptionValue[ClearDataBase] === Global`Only),
     Status["Deleting SQL data table: ", OptionValue[TableName]];
     DropSQLLesHouchesTable[TradeOptions[ObtainLesHouchesSpectrum, DropSQLLesHouchesTable, opt]];
     If[(OptionValue[ClearDataBase] === Global`Only),
         Return["Done"];
     ];
   ];
(*    SLAMPrint["1 T=",SessionTime[]-StartingSessionTime]; *)
   (* directly generating SQLCheckLiust (fast) *)
   If[Length[OptionValue[Model]]=!=0 || Head[OptionValue[Model]]=!=String,
     (* This is the case of direct model input *)
      (* Very simple consistency check... *)
      If[ModelGotNoModselQ[OptionValue[Model]],
      Message[SLAM::NoModselInModel];Abort[];
      ];
     (* SecureEntryFormats-> True  makes the Input save vs rounding errors. This is only for direct model input*)
     GeneratorInputSQLDefinitions = ToSQLLesHouchesData[OptionValue[Model], Direction -> In, Output -> Definitions, DataStructure -> Global`LesHouchesInputSQLFormats, NoComments -> True,SecureEntryFormats->True];
     ,
     (* Predifined Models use this:*)
     GeneratorInputSQLDefinitions = GenerateInputSQLChecklist[TradeOptions[ObtainLesHouchesSpectrum, GenerateInputSQLChecklist,opt]];
     (* Global`MyGeneratorInputSQLDefinitionsI=GeneratorInputSQLDefinitions;*)
(*     SLAMPrint["1a T=",SessionTime[]-StartingSessionTime]; *)

   ];
   (* Trying to extract Spectrum with matching InputDefinitions from Data Base *)
(*    SLAMPrint["2 T=",SessionTime[]-StartingSessionTime]; *)
   If[OptionValue[UseDataBase],
      Off[DatabaseLink`JDBC::"error"];

      SQLDataBaseOutput = SelectSQLLesHouchesTable[OutputFormat -> LesHouchesBlockRule, Direction -> Out, CheckList -> GeneratorInputSQLDefinitions, TradeOptions[ObtainLesHouchesSpectrum, SelectSQLLesHouchesTable, opt]];

      On[DatabaseLink`JDBC::"error"];      
      ,
      Status["No data base to be used."];
      SQLDataBaseOutput={};
   ];
(*    SLAMPrint["3 T=",SessionTime[]-StartingSessionTime]; *)


   If[SQLDataBaseOutput=== $Failed,
   (* Could not obtain values from table because it doesnt exist -> creating table *)
      Status["Creating new SQL table: ",OptionValue[TableName]];
      CreateSQLLesHouchesDataBase[TradeOptions[ObtainLesHouchesSpectrum, CreateSQLLesHouchesDataBase, opt]];
   ];
   Label["CALCANDSAVESPECTRUM"];

(*    SLAMPrint["4 T=",SessionTime[]-StartingSessionTime]; *)

   Which[Length[SQLDataBaseOutput] === 0,
         Status["No data match in data base!\nRunning spectrum generator..."];
    
         If[Length[OptionValue[Model]]=!=0,
            (* This is only for direct model input*)
            WriteLesHouchesFile[InputList ->(OptionValue[Model]), AppendToFile->False, TradeOptions[ObtainLesHouchesSpectrum, WriteLesHouchesFile, opt]];
            ,
           (* Predifined Models use this:*)
            WriteSpectrumGeneratorInput[TradeOptions[ObtainLesHouchesSpectrum, WriteSpectrumGeneratorInput, opt]];
         ];

(*          SLAMPrint["5 T=",SessionTime[]-StartingSessionTime]; *)

         (*Running spectrum generator *)
         Switch[{RunSpectrumGenerator[TradeOptions[ObtainLesHouchesSpectrum, RunSpectrumGenerator, opt]]},
                {},AllFine;,
                {$Aborted},Return[$Aborted];
               ];

(*          SLAMPrint["6 T=",SessionTime[]-StartingSessionTime]; *)

         (* Converting OutputFile to Mathematica BlockRule Notation*)
         NewSpectrumBlockRule = ReadLesHouchesSpectrumFile[Input -> File, InputRequest -> All, TradeOptions[ObtainLesHouchesSpectrum, ReadLesHouchesSpectrumFile, opt]];
         If[NewSpectrumBlockRule===$Aborted,Return[$Aborted];];
    
         If[OptionValue[ExtendDataBase]&&OptionValue[UseDataBase],
          (* Converting InputFile to Mathematica BlockRule Notation*)
     
            GeneratorInputBlockRule = ReadLesHouchesSpectrumFile[Input -> File, InputRequest -> All, OutputFilePath -> OptionValue[InputFilePath], TradeOptions[ObtainLesHouchesSpectrum, ReadLesHouchesSpectrumFile, opt]];

(*             SLAMPrint["7 T=",SessionTime[]-StartingSessionTime];  *)
           If[GeneratorInputBlockRule===$Aborted,Return[$Aborted];];
           (* Obtaining SQL Definitions of New Spectrum including comments*)

          
            GeneratorOutputSQLDefinitions = ToSQLLesHouchesData[NewSpectrumBlockRule, Direction -> Out, Output -> Definitions, DataStructure -> Global`LesHouchesOutputSQLFormats, NoComments -> False];
(*             SLAMPrint["8 T=",SessionTime[]-StartingSessionTime]; *)
     
           (* Obtaining SQL Definitions of Input values for Spectrum including comments*)
            GeneratorInputSQLDefinitions = ToSQLLesHouchesData[GeneratorInputBlockRule, Direction -> In, Output -> Definitions, DataStructure -> Global`LesHouchesInputSQLFormats, NoComments -> False];
(*             SLAMPrint["9 T=",SessionTime[]-StartingSessionTime]; *)
               (* Global`MyGeneratorInputSQLDefinitions=GeneratorInputSQLDefinitions; *)
            (* Adding new Spectrum to Data Base *)
            Status["Saving spectrum in data base: ",OptionValue[SQLDataBase]];
            WriteSQLLesHouchesData[TableDefinitionList -> Join[GeneratorInputSQLDefinitions, GeneratorOutputSQLDefinitions], TradeOptions[ObtainLesHouchesSpectrum, WriteSQLLesHouchesData, opt]];
(*             SLAMPrint["10 T=",SessionTime[]-StartingSessionTime]; *)
           ];
    
        Return[FilterLesHouchesBlockToInputRequest[NewSpectrumBlockRule, AdoptEntryFormats -> True, InputRequest -> OptionValue[InputRequest]]];
        ,
        Length[SQLDataBaseOutput] === 1,
        Status["Matching data found in data base..."];
        If[OptionValue[RefreshDataBase],
           Status["Deleting data base entry..."];
           DeleteSQLLesHouchesTable[Condition -> And @@ (DatabaseLink`SQLColumn[#[[1]]] == #[[2]] & /@ (GeneratorInputSQLDefinitions)),TradeOptions[ObtainLesHouchesSpectrum, DeleteSQLLesHouchesTable, opt]];
           SQLDataBaseOutput={};
           Goto["CALCANDSAVESPECTRUM"];
        ];
        Return[FilterLesHouchesBlockToInputRequest[SQLDataBaseOutput[[1]], AdoptEntryFormats -> True, InputRequest -> OptionValue[InputRequest]]];,
        Length[SQLDataBaseOutput] > 1,
        Message[SLAM::MultipleDBEntry];
        If[OptionValue[RefreshDataBase],
           Status["Deleting data base entries..."];
           DeleteSQLLesHouchesTable[Condition -> And @@ (DatabaseLink`SQLColumn[#[[1]]] == #[[2]] & /@ (GeneratorInputSQLDefinitions)),TradeOptions[ObtainLesHouchesSpectrum, DeleteSQLLesHouchesTable, opt]];
           SQLDataBaseOutput={};
           Goto["CALCANDSAVESPECTRUM"];
        ];
        Return[FilterLesHouchesBlockToInputRequest[SQLDataBaseOutput[[1]], AdoptEntryFormats -> True, InputRequest -> OptionValue[InputRequest]]];
   ];
   
];

(* Analyzing and writing FunctionSettings of OSQLR *)


Options[OSQLRAnalyzeFunction] = {
  ArgumentTypes -> {1 -> SQLDataTypeName[Real], 2 -> SQLDataTypeName[Real], 3 -> SQLDataTypeName[Real], Default -> SQLDataTypeName[Real]},
  ResultTypes -> {1 -> SQLDataTypeName[Real], 2 -> SQLDataTypeName[Real], 3 -> SQLDataTypeName[Real], Default -> SQLDataTypeName[Real]},
  ResultStructure -> 1};
OSQLRAnalyzeFunction[functionhead_, opt : OptionsPattern[]] := Module[{ListOfPatterns, ArgsCountings, InputSQLDefinition, InputSQLDeclaration,OutputSQLDefinition, OutputSQLDeclaration, ResPattern, resultpositions, resultnumberslist},
  (*(#[[1]]/.functionhead:>List)&/@ *)
  
  SLAMPrint["QRSQLAnalyzeFunction: Analyzing function ", functionhead, "..."];
  
  ListOfPatterns = (((#[[1]] /. {functionhead -> List, Pattern -> Pat}) & /@ DownValues[functionhead]) /. HoldPattern -> Join);
  
  ArgsCountings = MergeEquivalentRulesLists[({NumberOfBlankNullSequence -> Length[Cases[#, BlankNullSequence, {0, Infinity}, Heads -> True]],
                                              NumberOfOptArgs -> Length[Cases[#, OptionsPattern, {0, Infinity}, Heads -> True]],
                                              NumberOfArgs -> Length[# /. OptionsPattern -> OptionalPattern /. {Pat[_, OptionalPattern[]] :> Seq[]} /. Seq -> Sequence]} & /@ ListOfPatterns), 
                                            Operation -> Max];
  
  (* BlankNullSequence cannot be treated by OSQLR function !*)
  If[(NumberOfBlankNullSequence /. ArgsCountings) =!= 0,
   SLAMPrint["WARNING: OSQLR cannot treat functions using BlankNullSequence pattern!"];
   Return[$Aborted];];
  
  (* Processing normal arguments *)
  If[(NumberOfArgs /. ArgsCountings) > 0,
     InputSQLDeclaration = Table[{"ARGUMENT_" <> ToString[i], (i /. OptionValue[ArgumentTypes]) /. n_Integer :> (Default /. OptionValue[ArgumentTypes])}, {i, 1, NumberOfArgs /. ArgsCountings}];
     InputSQLDefinition = Table[{"ARGUMENT_" <> ToString[i], OSQLRArgument[i]}, {i, 1, NumberOfArgs /. ArgsCountings}];,
     InputSQLDeclaration = {};
     InputSQLDefinition = {};
   ];
  
  
  (* Checking if OptionPattern[] is used and Creating SQL Data with the help of options *)
  If[((NumberOfOptArgs /. ArgsCountings) =!= 0) && (Head[Options[functionhead]] === List),
     (AppendTo[InputSQLDeclaration, (({ToUpperCase[ToString[#[[1]]]] <> "_1",DetectDataType[#[[2]]]}))]; 
      AppendTo[InputSQLDefinition, (({ToUpperCase[ToString[#[[1]]]] <> "_1", #[[1]]}))];) & /@ Options[functionhead];
   ];
  
  If[! FreeQ[InputSQLDeclaration, $UnknownDataType],
     SLAMPrint["WARNING: Unknown data type in function argument!"]; 
     Return[$Aborted];];
  
  (* Processing Output formats*)
  resultnumberslist = Cases[OptionValue[ResultStructure], n_Integer, {0, Infinity}, Heads -> True];
  
  OutputSQLDeclaration = {"RESULT_" <> ToString[#], (# /. OptionValue[ResultTypes]) /. n_Integer :> (Default /. OptionValue[ResultTypes])} & /@ resultnumberslist;
  OutputSQLDefinition = {"RESULT_" <> ToString[#], OSQLRResult[#]} & /@resultnumberslist;
  
  ResPattern = OptionValue[ResultStructure] /. ((# -> OSQLRResult[#]) & /@ resultnumberslist);
  
  resultpositions = (OSQLRResult[#] -> Position[ResPattern, OSQLRResult[#]]) & /@ resultnumberslist;
  
  SLAMPrint["Adding function properties to OSQLR Options"];
  
  SetOptions[OSQLR, FunctionSettings -> Append[FunctionSettings /. Options[OSQLR], 
     ToString[functionhead] -> {TableName -> ToUpperCase[ToString[functionhead]], 
                                ArgumentProperties -> ArgsCountings, 
                                InputDeclaration -> InputSQLDeclaration, 
                                InputDefinition -> InputSQLDefinition, 
                                OutputDeclaration -> OutputSQLDeclaration, 
                                OutputDefinition -> OutputSQLDefinition, 
                                ResultPattern -> ResPattern, ResultPosition -> resultpositions,
                                ResultNumberList -> resultnumberslist}]];
  ];

(* Calculating and storing result *)

Options[OSQLRCalcFunction] = Union[{ExtendDataBase -> True}, Options[WriteSQLLesHouchesData]];
OSQLRCalcFunction[functionhead_, argslist_List, currentinputdefinition_, opt : OptionsPattern[]] := Module[{CalculationResult, ResultRule, functionname, n},
   
   CalculationResult = functionhead @@ argslist;
   
   functionname = ToString[functionhead];
   
   ResultRule = (OSQLRResult[#] -> Part[CalculationResult, ((OSQLRResult[#] /. (ResultPosition/. (functionname /. (FunctionSettings /. Options[OSQLR]))))) /. List -> Sequence]) & /@ (ResultNumberList /. (functionname /. (FunctionSettings /. Options[OSQLR])));
   
   
   If[ReplacePart[CalculationResult, (ResultPosition /. (functionname /.(FunctionSettings /. Options[OSQLR]))) /. Rule[a_, b_] :> Rule[b, a]] =!= (ResultPattern /. (functionname /. (FunctionSettings /. Options[OSQLR]))),
      Message[SLAM::NoResults,functionname @@ argslist];
      Return[$Failed]];
   
   (* Checking if result data type fits to saved result data types *)
      n = 1;
      If[! And @@ (If[((OutputDeclaration /. (functionname /. (FunctionSettings /. Options[OSQLR])))[[n++, 2]]) == DetectDataType[(OSQLRResult[#] /. ResultRule)], True, False] & /@ (ResultNumberList /. (functionname /.(FunctionSettings /. Options[OSQLR])))), 
         Message[SLAM::WrongResultFormat,functionname]; 
         Return[$Failed];];
   
(*    SLAMPrint[Join[currentinputdefinition, (OutputDefinition /. (functionname/. (FunctionSettings /. Options[OSQLR]))) /. ResultRule]]; *)
   
   (* Writing result to data base *)
   
   If[OptionValue[ExtendDataBase],
      SLAMPrint["Extending data base: ",OptionValue[SQLDataBase]];
      WriteSQLLesHouchesData[
        TableName -> (TableName /. (functionname /. (FunctionSettings /. Options[OSQLR]))),
        TableDefinitionList -> Join[currentinputdefinition, (OutputDefinition /. (functionname/. (FunctionSettings /. Options[OSQLR]))) /. ResultRule],
        TradeOptions[OSQLRCalcFunction, WriteSQLLesHouchesData, opt]];
    ];
   
   Return[(ResultPattern /. (functionname /. (FunctionSettings /. Options[OSQLR]))) /. ResultRule];
   
   ];


DetectDataType[data_] :=  Which[Head[data] === Real, 
                                 SQLDataTypeName[Real], 
                                Head[data] === String, 
                                 SQLDataTypeName[String], 
                                Head[data] === Integer, 
                                 SQLDataTypeName[Integer], 
                                True, 
                                 $UnknownDataType]

(* Is used to combine equivalent list of rules in a list to a single list of rule*)
(* example: 
   MergeEquivalentRulesLists[{{test->1,test2->2},{test->3,test2->0}},Operation -> Max] 
   -> 
   {test->3,test2->2}  
*)
Options[MergeEquivalentRulesLists] = {Operation -> Max};
MergeEquivalentRulesLists[RulesList_List, opt : OptionsPattern[]] := Module[{LeftRule, n},
  LeftRule = RulesList[[1]] /. Rule[x_, y_] :> x;
  n = 1;
  Return[(LeftRule[[n++]] -> #) & /@ (OptionValue[Operation][#] & /@ Transpose[((#[[2]]) & /@ #) & /@ RulesList])];
  ]



Options[OSQLR] = {
   SQLDataBase -> Global`SQLDataBaseValue,
   SQLDataBaseOptions -> Global`SQLDataBaseOptionsValue,
   SQLDataBaseType -> Global`SQLDataBaseTypeValue,
   FunctionSettings -> {},
   UseDataBase->True,
   ExtendDataBase -> True,
   RefreshDataBase -> False,
   ClearFunctionStorage -> False,
   ArgumentTypes -> {1 -> SQLDataTypeName[Real], 2 -> SQLDataTypeName[Real], 3 -> SQLDataTypeName[Real], Default -> SQLDataTypeName[Real]},
   ResultTypes -> {1 -> SQLDataTypeName[Real], 2 -> SQLDataTypeName[Real], 3 -> SQLDataTypeName[Real], Default -> SQLDataTypeName[Real]},
   ResultStructure -> 1};


(* wraps a function in order to store calculated results in sql data base *)
OSQLR[functionhead_, opt : OptionsPattern[]][a___] := OSQLR[functionhead, {a}, opt]
OSQLR[functionhead_, argslist_List, opt : OptionsPattern[]] := Module[{functionsettings, functionname, currentinputdefinition, n, argrule, optargrule, RawResultList, CalculationResult, ResultRule},
  
  If[OptionValue[UseDataBase],
  functionname = ToString[functionhead];
  
  (*Checking if function is already known*)
  
  If[Length[functionsettings = (functionname /. OptionValue[FunctionSettings])] === 0,
   (* Unknown functions need to be analyzed*)
      OSQLRAnalyzeFunction[functionhead, TradeOptions[OSQLR, OSQLRAnalyzeFunction, opt]];
      functionsettings = (functionname /. OptionValue[FunctionSettings]);
   ];
  
  
  If[(OptionValue[ClearFunctionStorage] === True) || (OptionValue[ClearFunctionStorage] === Global`Only),
     SLAMPrint["Deleting data storage of function: ", functionname];
     DropSQLLesHouchesTable[TableName -> (TableName /. (functionname /. OptionValue[FunctionSettings])),
                            TradeOptions[OSQLR, SelectSQLLesHouchesTable, opt]];
  If[(OptionValue[ClearFunctionStorage] === Global`Only),
    (*Removing rule for current function in FunctionSettings OptionList*)
     SetOptions[OSQLR, FunctionSettings -> Complement[#, FilterRules[#, functionname]] &@(FunctionSettings /. Options[OSQLR])];
     Return["Done"];
    ];
   ];
  
  (* Determination of current input *)
  {argrule, optargrule} = {(n = 1; (OSQLRArgument[n++] -> # & /@ (#))), Complement[argslist, #]} &@(argslist /. Rule[a_, b_] -> Sequence[]);
(*   SLAMPrint[{argrule, optargrule}]; *)
  currentinputdefinition = ((InputDefinition /. (functionname /. OptionValue[FunctionSettings])) /. argrule /. optargrule /. Options[functionhead]) /. List[x_, OSQLRArgument[n_]] :> Sequence[];
  
(*   SLAMPrint[currentinputdefinition]; *)
  
  
  (* Getting Results from SQL data base although there might not be any data ;)*)
  Off[DatabaseLink`JDBC::"error"];
  RawResultList = SelectSQLLesHouchesTable[CheckList -> currentinputdefinition,
                                           OutputRequestList -> (OutputDeclaration /. (functionname /. OptionValue[FunctionSettings])),
                                           OutputFormat -> PureValueList,
                                           TableName -> (TableName /. (functionname /. OptionValue[FunctionSettings])),
                                           TradeOptions[OSQLR, SelectSQLLesHouchesTable, opt]];
  On[DatabaseLink`JDBC::"error"];
  
  If[RawResultList === $Failed,
   (*Table seems not to be declared*)
     SLAMPrint["Creating data base table: ", (TableName /. (functionname /. OptionValue[FunctionSettings])), "..."];
     CreateSQLLesHouchesDataBase[TableName -> (TableName /. (functionname /. OptionValue[FunctionSettings])),
                                 TableDeclarationList -> Join[InputDeclaration /. (functionname /. OptionValue[FunctionSettings]), OutputDeclaration /. (functionname /. OptionValue[FunctionSettings])],
                                 TradeOptions[OSQLR, CreateSQLLesHouchesDataBase, opt]];
                                 RawResultList = {};
   ];
  
  If[RawResultList == {},
   (*Nothing found in data base*)
   SLAMPrint["No result in data base! Starting calculation... "];
   
   Return[OSQLRCalcFunction[functionhead, 
                            argslist, 
                            currentinputdefinition, 
                            TradeOptions[OSQLR, OSQLRCalcFunction, opt]]];
   ];
  
  
  If[Length[RawResultList] > 0,
     If[OptionValue[RefreshDataBase],
        SLAMPrint["Refreshing matching result in data base..."];
        DeleteSQLLesHouchesTable[TableName -> (TableName /. (functionname /. OptionValue[FunctionSettings])),
        Condition -> And @@ (DatabaseLink`SQLColumn[#[[1]]] == #[[2]] & /@ (currentinputdefinition))
      ];
     Return[OSQLRCalcFunction[functionhead, 
                              argslist, 
                              currentinputdefinition, 
                              TradeOptions[OSQLR, OSQLRCalcFunction, opt]]];
     ,
     SLAMPrint["Found matching result in data base."];
     n = 1;
     ResultRule = (OSQLRResult[#] -> RawResultList[[n++]]) & /@ (ResultNumberList /. (functionname /. (FunctionSettings /. Options[OSQLR])));
     Return[(ResultPattern /. (functionname /. (FunctionSettings /. Options[OSQLR]))) /. ResultRule];
     ];
   ];
   ,(*When not using data base simply return the requested function with proper arguments*)
   Return[functionhead[Sequence@@argslist]];
  ];
  ];



(* # END OF SQL Data Base Interface # *)
End[];
EndPackage[];


SLAMPrint["SLAM v1.0 loaded."];

